import unittest
from libpysal.examples import load_example
import geopandas as gpd
import numpy as np
from segregation.multigroup import MultiInfoTheory


class Multi_Information_Theory_Tester(unittest.TestCase):
    def test_Multi_Information_Theory(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        groups_list = ['WHITE', 'BLACK', 'ASIAN','HISP']
        df = s_map[groups_list]
        index = MultiInfoTheory(df, groups_list)
        np.testing.assert_almost_equal(index.statistic, 0.1710160297858887)


if __name__ == '__main__':
    unittest.main()


"""Batch compute wrappers for calculating all relevant statistics at once."""

import inspect
import warnings

import pandas as pd
from tqdm.auto import tqdm

from .. import multigroup, singlegroup
from .._base import SpatialImplicitIndex
from ..dynamics import compute_multiscalar_profile

singlegroup_classes = {}
for name, obj in inspect.getmembers(singlegroup):
    if inspect.isclass(obj):
        singlegroup_classes[name] = obj

multigroup_classes = {}
for name, obj in inspect.getmembers(multigroup):
    if inspect.isclass(obj):
        multigroup_classes[name] = obj

implicit_single_indices = {}
for name, obj in inspect.getmembers(singlegroup):
    if inspect.isclass(obj):
        if str(SpatialImplicitIndex) in [str(i) for i in obj.__bases__]:
            implicit_single_indices[name] = obj

implicit_multi_indices = {}
for name, obj in inspect.getmembers(multigroup):
    if inspect.isclass(obj):
        if str(SpatialImplicitIndex) in [str(i) for i in obj.__bases__]:
            implicit_multi_indices[name] = obj


def batch_compute_singlegroup(
    gdf, group_pop_var, total_pop_var, progress_bar=True, **kwargs
):
    """Batch compute single-group indices.

    Parameters
    ----------
    gdf : DataFrame or GeoDataFrame
        DataFrame holding demographic data for study region
    group_pop_var : str
        The name of variable in data that contains the population size of the group of interest
    total_pop_var : str
        Variable in data that contains the total population count of the unit
    progress_bar: bool
        Whether to show a progress bar during calculation
    **kwargs : dict
        additional keyword arguments passed to each index (e.g. for setting a random
        seed in indices like ModifiedGini or ModifiedDissm)

    Returns
    -------
    pandas.DataFrame
        dataframe with statistic name as dataframe index and statistic value as dataframe values
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        fitted = {}
        if progress_bar:
            pbar = tqdm(total=len(singlegroup_classes.keys()))

            for each in sorted(singlegroup_classes.keys()):
                pbar.set_description(each)
                fitted[each] = singlegroup_classes[each](
                    gdf, group_pop_var, total_pop_var, **kwargs
                ).statistic
                pbar.update(1)
        else:
            for each in sorted(singlegroup_classes.keys()):
                fitted[each] = singlegroup_classes[each](
                    gdf, group_pop_var, total_pop_var, **kwargs
                ).statistic
        fitted = pd.DataFrame.from_dict(fitted, orient="index").round(4)
        fitted.columns = ["Statistic"]
        fitted.index.name = "Name"
        return fitted


def batch_compute_multigroup(gdf, groups, **kwargs):
    """Batch compute multi-group indices.

    Parameters
    ----------
    gdf : DataFrame or GeoDataFrame
        DataFrame holding demographic data for study region
    groups : list
        The variables names in data of the groups of interest of the analysis.
    **kwargs : dict
        additional keyword arguments passed to each index (e.g. for setting a random
        seed in indices like ModifiedGini or ModifiedDissm)

    Returns
    -------
    pandas.DataFrame
        dataframe with statistic name as dataframe index and statistic value as dataframe values
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        fitted = {}
        for each in sorted(multigroup_classes.keys()):
            fitted[each] = multigroup_classes[each](gdf, groups, **kwargs).statistic
        fitted = pd.DataFrame.from_dict(fitted, orient="index").round(4)
        fitted.columns = ["Statistic"]
        fitted.index.name = "Name"
    return fitted


def batch_multiscalar_singlegroup(
    gdf, distances, group_pop_var, total_pop_var, progress_bar=True, **kwargs
):
    """Batch compute multiscalar profiles for single-group indices.

    Parameters
    ----------
    gdf : DataFrame or GeoDataFrame
        DataFrame holding demographic data for study region
    distances : list
        list of floats representing bandwidth distances that define a local
        environment.
    group_pop_var : str
        The name of variable in data that contains the population size of the group
        of interest
    total_pop_var : str
        Variable in data that contains the total population count of the unit
    progress_bar: bool
        Whether to show a progress bar during calculation
    **kwargs : dict
        additional keyword arguments passed to each index (e.g. for setting a random
        seed in indices like ModifiedGini or ModifiedDissm)

    Returns
    -------
    pandas.DataFrame
        pandas Dataframe with distance as dataframe index and each segregation
        statistic as dataframe columns
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        profs = []
        if progress_bar:
            pbar = tqdm(total=len(implicit_single_indices.keys()))
            for idx in sorted(implicit_single_indices.keys()):
                pbar.set_description(idx)
                prof = compute_multiscalar_profile(
                    gdf=gdf,
                    segregation_index=implicit_single_indices[idx],
                    distances=distances,
                    group_pop_var=group_pop_var,
                    total_pop_var=total_pop_var,
                    **kwargs
                )
                profs.append(prof)
                pbar.update(1)
        else:
            for idx in sorted(implicit_single_indices.keys()):
                prof = compute_multiscalar_profile(
                    gdf=gdf,
                    segregation_index=implicit_single_indices[idx],
                    distances=distances,
                    group_pop_var=group_pop_var,
                    total_pop_var=total_pop_var,
                    **kwargs
                )
                profs.append(prof)
        df = pd.concat(profs, axis=1)
        return df


def batch_multiscalar_multigroup(gdf, distances, groups, progress_bar=True, **kwargs):
    """Batch compute multiscalar profiles for multi-group indices.

    Parameters
    ----------
    gdf : DataFrame or GeoDataFrame
        DataFrame holding demographic data for study region
    distances : list
        list of floats representing bandwidth distances that define a local
        environment.
    groups : list
        The variables names in data of the groups of interest of the analysis.
    progress_bar: bool
        Whether to show a progress bar during calculation
    **kwargs : dict
        additional keyword arguments passed to each index (e.g. for setting a random
        seed in indices like ModifiedGini or ModifiedDissm)

    Returns
    -------
    pandas.DataFrame
        pandas Dataframe with distance as dataframe index and each segregation
        statistic as dataframe columns
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        profs = []
        if progress_bar:
            pbar = tqdm(total=len(implicit_multi_indices.keys()))
            for idx in sorted(implicit_multi_indices.keys()):
                pbar.set_description(idx)
                prof = compute_multiscalar_profile(
                    gdf=gdf,
                    segregation_index=implicit_multi_indices[idx],
                    distances=distances,
                    groups=groups,
                    **kwargs
                )
                profs.append(prof)
                pbar.update(1)

        else:
            for idx in sorted(implicit_multi_indices.keys()):
                prof = compute_multiscalar_profile(
                    gdf=gdf,
                    segregation_index=implicit_multi_indices[idx],
                    distances=distances,
                    groups=groups,
                    **kwargs
                )
                profs.append(prof)
        df = pd.concat(profs, axis=1)
        return df



import unittest
import pandas as pd
import geopandas as gpd
import numpy as np
from libpysal.examples import load_example
from segregation.inference import SingleValueTest, TwoValueTest
from segregation.multigroup import MultiDissim
from segregation.singlegroup import Dissim


class Inference_Tester(unittest.TestCase):
    def test_Inference(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        # note need to recast as datafrme
        s_map_no_geom = pd.DataFrame(
            gpd.read_file(
                load_example("Sacramento1").get_path("sacramentot2.shp")
            ).drop(columns=["geometry"])
        )
        index1 = Dissim(s_map, "HISP", "TOT_POP")
        index2 = Dissim(s_map, "BLACK", "TOT_POP")

        groups_list = ["WHITE", "BLACK", "ASIAN", "HISP"]
        m_index = MultiDissim(s_map, groups_list)

        m_index_1 = MultiDissim(s_map[0:200], groups_list)
        m_index_2 = MultiDissim(s_map[200:], groups_list)

        # Single Value Tests #
        np.random.seed(123)
        res = SingleValueTest(
            index1,
            null_approach="systematic",
            iterations_under_null=50,
            backend="multiprocessing",
        )
        np.testing.assert_almost_equal(res.est_sim.mean(), 0.017621, decimal=2)

        index1_no_geom = Dissim(s_map_no_geom, "HISP", "TOT_POP")

        np.random.seed(123)
        res = SingleValueTest(
            index1_no_geom,
            null_approach="systematic",
            iterations_under_null=50,
            backend="multiprocessing",
        )
        np.testing.assert_almost_equal(res.est_sim.mean(), 0.017621, decimal=2)

        np.random.seed(123)
        res = SingleValueTest(
            index1, null_approach="bootstrap", iterations_under_null=50
        )
        np.testing.assert_almost_equal(res.est_sim.mean().round(2), 0.32, decimal=2)

        np.random.seed(123)
        res = SingleValueTest(
            index1, null_approach="evenness", iterations_under_null=50
        )
        np.testing.assert_almost_equal(
            res.est_sim.mean(), 0.01596295861644252, decimal=2
        )

        np.random.seed(123)
        res = SingleValueTest(
            index1, null_approach="geographic_permutation", iterations_under_null=50
        )
        np.testing.assert_almost_equal(
            res.est_sim.mean(), 0.32184656076566864, decimal=2
        )

        np.random.seed(123)
        res = SingleValueTest(
            index1, null_approach="systematic_permutation", iterations_under_null=50
        )
        np.testing.assert_almost_equal(res.est_sim.mean().round(2), 0.016, decimal=2)

        np.random.seed(123)
        res = SingleValueTest(
            index1, null_approach="even_permutation", iterations_under_null=50
        )
        np.testing.assert_almost_equal(
            res.est_sim.mean(), 0.01619436868061094, decimal=2
        )

        np.random.seed(123)
        res = SingleValueTest(
            m_index, null_approach="bootstrap", iterations_under_null=50
        )
        np.testing.assert_almost_equal(res.est_sim.mean().round(2), 0.41, decimal=2)

        np.random.seed(123)
        res = SingleValueTest(
            m_index, null_approach="evenness", iterations_under_null=50
        )
        np.testing.assert_almost_equal(
            res.est_sim.mean(), 0.01633979237418177, decimal=2
        )

        np.random.seed(123)
        res = SingleValueTest(
            m_index, null_approach="person_permutation", iterations_under_null=50
        )
        np.testing.assert_almost_equal(
            res.est_sim.mean(), 0.01633979237418177, decimal=2
        )

        # Two Value Tests #
        np.random.seed(123)
        res = TwoValueTest(
            index1,
            index2,
            null_approach="random_label",
            iterations_under_null=50,
            backend="multiprocessing",
        )
        np.testing.assert_almost_equal(
            res.est_sim.mean(), -0.0031386146371949076, decimal=2
        )

        np.random.seed(123)
        res = TwoValueTest(
            index1,
            index2,
            null_approach="composition",
            iterations_under_null=50,
        )
        np.testing.assert_almost_equal(
            res.est_sim.mean(), -0.005032145622504718, decimal=2
        )

        np.random.seed(123)
        res = TwoValueTest(
            index1,
            index2,
            null_approach="share",
            iterations_under_null=50,
        )
        np.testing.assert_almost_equal(
            res.est_sim.mean(), -0.034350440515125, decimal=2
        )

        np.random.seed(123)
        res = TwoValueTest(
            index1,
            index2,
            null_approach="dual_composition",
            iterations_under_null=50,
        )
        np.testing.assert_almost_equal(
            res.est_sim.mean(), -0.004771386292706747, decimal=2
        )

        np.random.seed(123)
        res = TwoValueTest(
            m_index_1, m_index_2, null_approach="random_label", iterations_under_null=50
        )
        np.testing.assert_almost_equal(
            res.est_sim.mean(), -0.0024327144012562685, decimal=2
        )

        np.random.seed(123)
        res = TwoValueTest(
            m_index_1, m_index_2, null_approach="bootstrap", iterations_under_null=50
        )
        np.testing.assert_almost_equal(res.est_sim[0].mean(), 0.38738, decimal=2)

        np.random.seed(123)
        res = TwoValueTest(
            m_index_1,
            m_index_2,
            null_approach="person_permutation",
            iterations_under_null=50,
        )
        np.testing.assert_almost_equal(res.est_sim.mean(), 0.0, decimal=2)


if __name__ == "__main__":
    unittest.main()



"""Distance Decay Isolation Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
import pandas as pd
from ..util import generate_distance_matrix

from .._base import SingleGroupIndex, SpatialExplicitIndex


def _distance_decay_isolation(data, group_pop_var, total_pop_var, alpha=0.6, beta=0.5):
    """Calculate of Distance Decay Isolation index.

    Parameters
    ----------
    data          : a geopandas DataFrame with a geometry column.
    group_pop_var : string
                    The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
                    The name of variable in data that contains the total population of the unit
    alpha         : float
                    A parameter that estimates the extent of the proximity within the same unit. Default value is 0.6
    beta          : float
                    A parameter that estimates the extent of the proximity within the same unit. Default value is 0.5

    Returns
    ----------
    statistic : float
                Distance Decay Isolation Index
    core_data : a geopandas DataFrame
                A geopandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    It may be interpreted as the probability that the next person a group member meets anywhere in space is from the same group.

    Based on Morgan, Barrie S. "A distance-decay based interaction index to measure residential segregation." Area (1983): 211-217.

    The pairwise distance between unit i and itself is (alpha * area_of_unit_i) ^ beta.

    Reference: :cite:`morgan1983distance`.

    """
    if alpha < 0:
        raise ValueError("alpha must be greater than zero.")

    if beta < 0:
        raise ValueError("beta must be greater than zero.")

    x = np.array(data[group_pop_var])
    t = np.array(data[total_pop_var])

    if any(t < x):
        raise ValueError(
            "Group of interest population must equal or lower than the total population of the units."
        )

    X = x.sum()

    dist = generate_distance_matrix(data)

    np.fill_diagonal(dist, val=np.exp(-((alpha * data.area.values) ** (beta))))

    c = 1 - dist.copy()  # proximity matrix

    Pij = np.multiply(c, t) / np.sum(np.multiply(c, t), axis=1)

    DDxPx = (
        np.array(x / X) * np.nansum(np.multiply(Pij, np.array(x / t)), axis=1)
    ).sum()

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return DDxPx, core_data


class DistanceDecayIsolation(SingleGroupIndex, SpatialExplicitIndex):
    """Distance-Decay Isolation Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    alpha  : float
        A parameter that estimates the extent of the proximity within the same unit. Default value is 0.6
    beta : float
        A parameter that estimates the extent of the proximity within the same unit. Default value is 0.5

    Attributes
    ----------
    statistic : float
        Distance Decay Isolation Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    It may be interpreted as the probability that the next person a group member meets anywhere in space is from the same group.

    Based on Morgan, Barrie S. "A distance-decay based interaction index to measure residential segregation." Area (1983): 211-217.

    The pairwise distance between unit i and itself is (alpha * area_of_unit_i) ^ beta.

    Reference: :cite:`morgan1983distance`.
    """

    def __init__(
        self, data, group_pop_var, total_pop_var, alpha=0.6, beta=0.5, **kwargs,
    ):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        SpatialExplicitIndex.__init__(self,)
        self.alpha = alpha
        self.beta = beta
        aux = _distance_decay_isolation(
            self.data, self.group_pop_var, self.total_pop_var, self.alpha, self.beta,
        )

        self.statistic = aux[0]
        self.core_data = aux[1]
        self._function = _distance_decay_isolation



import unittest
from libpysal.examples import load_example
import geopandas as gpd
import numpy as np
from segregation.singlegroup import Dissim
from segregation.decomposition import DecomposeSegregation


class Decomposition_Tester(unittest.TestCase):
    def test_Decomposition(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        index1 = Dissim(s_map, 'HISP', 'TOT_POP')
        index2 = Dissim(s_map, 'BLACK', 'TOT_POP')
        res = DecomposeSegregation(index1, index2, counterfactual_approach = "composition")
        np.testing.assert_almost_equal(res.c_a, -0.16138819842911295)
        np.testing.assert_almost_equal(res.c_s, -0.005104643275796905)
        res.plot(plot_type = 'cdfs')
        res.plot(plot_type = 'maps')
        
        res = DecomposeSegregation(index1, index2, counterfactual_approach = "share")
        np.testing.assert_almost_equal(res.c_a, -0.1543828579279878)
        np.testing.assert_almost_equal(res.c_s, -0.012109983776922045)
        res.plot(plot_type = 'cdfs')
        res.plot(plot_type = 'maps')
        
        res = DecomposeSegregation(index1, index2, counterfactual_approach = "dual_composition")
        np.testing.assert_almost_equal(res.c_a, -0.16159526946235048)
        np.testing.assert_almost_equal(res.c_s, -0.004897572242559378)
        res.plot(plot_type = 'cdfs')
        res.plot(plot_type = 'maps')

if __name__ == '__main__':
    unittest.main()



"""Multigroup Relative Centralization index"""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import libpysal as lps
import numpy as np
from segregation.singlegroup import RelativeCentralization

from .._base import MultiGroupIndex, SpatialExplicitIndex

np.seterr(divide="ignore", invalid="ignore")


def _local_relative_centralization(data, group_pop_var, total_pop_var, W=None, k=5):
    """
    Calculation of Local Relative Centralization index for each unit

    Parameters
    ----------

    data          : a geopandas DataFrame with a geometry column.
    
    group_pop_var : string
                    The name of variable in data that contains the population size of the group of interest
                    
    total_pop_var : string
                    The name of variable in data that contains the total population of the unit
    
    k_neigh       : integer greater than 0. Default is 5.
                    Number of assumed neighbors for local context (it uses k-nearest algorithm method)
                    
    Returns
    -------

    statistics : np.array(n)
                 Local Relative Centralization values for each unit
                
    core_data  : a pandas DataFrame
                 A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Folch, David C., and Sergio J. Rey. "The centralization index: A measure of local spatial segregation." Papers in Regional Science 95.3 (2016): 555-576.
    
    Reference: :cite:`folch2016centralization`.
    """

    data = data.copy()
    if not W:
        W = lps.weights.KNN.from_dataframe(data, k=5)

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    c_lons = data.centroid.map(lambda p: p.x)
    c_lats = data.centroid.map(lambda p: p.y)

    points = list(zip(c_lons, c_lats))
    kd = lps.cg.kdtree.KDTree(np.array(points))

    local_RCEs = np.empty(len(data))

    for i in range(len(data)):

        x = list(W.neighbors.values())[i]
        x.append(
            list(W.neighbors.keys())[i]
        )  # Append in the end the current unit i inside the local context

        local_data = data.iloc[x, :].copy()

        # The center is given by the last position (i.e. the current unit i)
        local_RCE = RelativeCentralization(
            local_data, group_pop_var, total_pop_var, center=len(local_data) - 1
        )

        local_RCEs[i] = local_RCE.statistic

    return local_RCEs, core_data


class LocalRelativeCentralization(MultiGroupIndex, SpatialExplicitIndex):
    """Multigroup Local Simpson's Concentration Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    w : libpysal.W, optional
        lipysal spatial weights object used to define a local neighborhood. If none is passed,
        a KNN ojbect with k=5 will be used
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
        Multigroup Dissimilarity Index value
    core_data : a pandas DataFrame
        DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Folch, David C., and Sergio J. Rey. "The centralization index: A measure of local spatial segregation." Papers in Regional Science 95.3 (2016): 555-576.
    
    Reference: :cite:`folch2016centralization`.
    """

    def __init__(
        self,
        data,
        group_pop_var=None,
        total_pop_var=None,
        w=None,
        network=None,
        distance=None,
        decay=None,
        precompute=None,
        groups=None,
    ):
        """Init."""

        MultiGroupIndex.__init__(self, data, groups)
        if any([w, network, distance]):
            SpatialExplicitIndex.__init__(self)
        aux = _local_relative_centralization(
            self.data, group_pop_var=group_pop_var, total_pop_var=total_pop_var, W=w
        )

        self.statistics = aux[0]
        self.data = aux[1]
        self._function = _local_relative_centralization



import unittest
from libpysal.examples import load_example
import geopandas as gpd
import numpy as np
from segregation.singlegroup import RelativeConcentration


class Relative_Concentration_Tester(unittest.TestCase):
    def test_Relative_Concentration(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        df = s_map[['geometry', 'HISP', 'TOT_POP']]
        index = RelativeConcentration(df, 'HISP', 'TOT_POP')
        np.testing.assert_almost_equal(index.statistic, 0.12733820870675222)


if __name__ == '__main__':
    unittest.main()



"""Spatial Proximity Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
from ..util import generate_distance_matrix

from .._base import SingleGroupIndex, SpatialExplicitIndex


def _spatial_proximity(data, group_pop_var, total_pop_var, alpha=0.6, beta=0.5):
    """Calculate Spatial Proximity index.

    Parameters
    ----------
    data          : a geopandas DataFrame with a geometry column.
    group_pop_var : string
                    The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
                    The name of variable in data that contains the total population of the unit
    alpha         : float
                    A parameter that estimates the extent of the proximity within the same unit. Default value is 0.6
    beta          : float
                    A parameter that estimates the extent of the proximity within the same unit. Default value is 0.5
    metric        : string. Can be 'euclidean' or 'haversine'. Default is 'euclidean'.
                    The metric used for the distance between spatial units.
                    If the projection of the CRS of the geopandas DataFrame field is in degrees, this should be set to 'haversine'.

    Returns
    ----------
    statistic : float
                Spatial Proximity Index
    core_data : a geopandas DataFrame
                A geopandas DataFrame that contains the columns used to perform the estimate.

   Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    The pairwise distance between unit i and itself is (alpha * area_of_unit_i) ^ beta.

    Reference: :cite:`massey1988dimensions`.

    """
    if alpha < 0:
        raise ValueError("alpha must be greater than zero.")

    if beta < 0:
        raise ValueError("beta must be greater than zero.")

    T = data[total_pop_var].sum()

    data = data.assign(
        xi=data[group_pop_var],
        yi=data[total_pop_var] - data[group_pop_var],
        ti=data[total_pop_var],
    )

    X = data.xi.sum()
    Y = data.yi.sum()

    dist = generate_distance_matrix(data)

    np.fill_diagonal(dist, val=np.exp(-((alpha * data.area.values) ** (beta))))

    c = 1 - dist.copy()  # proximity matrix

    Pxx = ((np.array(data.xi) * c).T * np.array(data.xi)).sum() / X ** 2
    Pyy = ((np.array(data.yi) * c).T * np.array(data.yi)).sum() / Y ** 2
    Ptt = ((np.array(data.ti) * c).T * np.array(data.ti)).sum() / T ** 2
    SP = (X * Pxx + Y * Pyy) / (T * Ptt)

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return SP, core_data


class SpatialProximity(SingleGroupIndex, SpatialExplicitIndex):
    """Spatial Proximity Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    alpha  : float
        A parameter that estimates the extent of the proximity within the same unit. Default value is 0.6
    beta : float
        A parameter that estimates the extent of the proximity within the same unit. Default value is 0.5
    metric : string. Can be 'euclidean' or 'haversine'. Default is 'euclidean'.
        The metric used for the distance between spatial units.
        If the projection of the CRS of the geopandas DataFrame field is in degrees, this should be set to 'haversine'.


    Attributes
    ----------
    statistic : float
        Spatial Proximity Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    The pairwise distance between unit i and itself is (alpha * area_of_unit_i) ^ beta.

    Reference: :cite:`massey1988dimensions`.
    """

    def __init__(
        self, data, group_pop_var, total_pop_var, alpha=0.6, beta=0.5, **kwargs,
    ):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        SpatialExplicitIndex.__init__(self,)
        self.alpha = alpha
        self.beta = beta
        aux = _spatial_proximity(
            self.data, self.group_pop_var, self.total_pop_var, self.alpha, self.beta,
        )

        self.statistic = aux[0]
        self.core_data = aux[1]
        self._function = _spatial_proximity



"""Multigroup dissimilarity index"""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np

from .._base import MultiGroupIndex, SpatialImplicitIndex

np.seterr(divide="ignore", invalid="ignore")


def _multi_location_quotient(data, groups):
    """
    Calculation of Location Quotient index for each group and unit

    Parameters
    ----------

    data   : a pandas DataFrame of n rows
    
    groups : list of strings of length k.
             The variables names in data of the groups of interest of the analysis.

    Returns
    -------

    statistics : np.array(n,k)
                 Location Quotient values for each group and unit.
                 Column k has the Location Quotient of position k in groups.
                
    core_data  : a pandas DataFrame
                 A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Isard, Walter. Methods of regional analysis. Vol. 4. Рипол Классик, 1967.
    
    Reference: :cite:`isard1967methods`.

    """
    
    core_data = data[groups]
    
    df = np.array(core_data)
    
    n = df.shape[0]
    K = df.shape[1]
    
    T = df.sum()
    
    ti = df.sum(axis = 1)
    Xk = df.sum(axis = 0)
    
    multi_LQ = (df / np.repeat(ti, K, axis = 0).reshape(n,K)) / (Xk / T)
    
    return multi_LQ, core_data, groups

        
class MultiLocationQuotient(MultiGroupIndex, SpatialImplicitIndex):
    """Multigroup Local Diversity Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
        Multigroup Dissimilarity Index value
    core_data : a pandas DataFrame
        DataFrame that contains the columns used to perform the estimate.

    Notes
    -----    
    Reference: :cite:`isard1967methods`.
    """

    def __init__(
        self,
        data,
        groups,
        w=None,
        network=None,
        distance=None,
        decay=None,
        precompute=None,
        function='triangular'
    ):
        """Init."""

        MultiGroupIndex.__init__(self, data, groups)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(self, w, network, distance, decay, function, precompute)
        aux = _multi_location_quotient(self.data, self.groups)

        self.statistics = aux[0]
        self.data = aux[1]
        self.groups = aux[2]
        self._function = _multi_location_quotient



import unittest
from libpysal.examples import load_example
import geopandas as gpd
import numpy as np
from segregation.singlegroup import DistanceDecayIsolation


class Distance_Decay_Isolation_Tester(unittest.TestCase):
    def test_Distance_Decay_Isolation(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        df = s_map[['geometry', 'HISP', 'TOT_POP']]
        df = df.to_crs(df.estimate_utm_crs())
        index = DistanceDecayIsolation(df, 'HISP', 'TOT_POP')
        np.testing.assert_almost_equal(index.statistic, 0.20144710843904595)


if __name__ == '__main__':
    unittest.main()



from libpysal.examples import load_example
import geopandas as gpd
import numpy as np
from segregation.multigroup import MultiDissim
from segregation.dynamics import compute_multiscalar_profile
import quilt3
import pandana as pdna


p = quilt3.Package.browse('osm/metro_networks_8k', "s3://spatial-ucr/")
p['40900.h5'].fetch()
net = pdna.Network.from_hdf5('40900.h5')

def test_multiscalar():
    s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
    df = s_map.to_crs(s_map.estimate_utm_crs())
    profile = compute_multiscalar_profile(
        gdf=df,
        segregation_index=MultiDissim,
        distances=[500, 1000, 1500, 2000],
        groups=["HISP", "BLACK", "WHITE"],
    )
    np.testing.assert_array_almost_equal(
        profile.values, [0.4246, 0.4246, 0.4173, 0.4008, 0.3776], decimal=4
    )


def test_multiscalar_network():
    s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
    df = s_map.to_crs(s_map.estimate_utm_crs())
    profile = compute_multiscalar_profile(
        gdf=df,
        segregation_index=MultiDissim,
        distances=[500, 1000],
        groups=["HISP", "BLACK", "WHITE"],
    )
    np.testing.assert_array_almost_equal(
        profile.values, [0.4247, 0.4246, 0.4173], decimal=4
    )



"""
:mod:`segregation` --- Spatial/Aspatial Segregation Analysis
=================================================

"""
import contextlib
from importlib.metadata import PackageNotFoundError, version

from . import batch, decomposition, dynamics, inference, local, multigroup, singlegroup, util

with contextlib.suppress(PackageNotFoundError):
    __version__ = version("segregation")


"""Perimeter-Area Ratio Dissimilarity Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
from sklearn.metrics.pairwise import manhattan_distances

from .._base import SingleGroupIndex, SpatialExplicitIndex, _return_length_weighted_w
from .dissim import _dissim


def _perimeter_area_ratio_spatial_dissim(
    data, group_pop_var, total_pop_var, standardize=True
):
    """Calculation of Perimeter/Area Ratio Spatial Dissimilarity index.

    Parameters
    ----------
    data          : a geopandas DataFrame with a geometry column.
    group_pop_var : string
                    The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
                    The name of variable in data that contains the total population of the unit
    standardize   : boolean
                    A condition for standardisation of the weights matrices.
                    If True, the values of cij in the formulas gets standardized and the overall sum is 1.

    Returns
    ----------
    statistic : float
                Perimeter/Area Ratio Spatial Dissimilarity Index
    core_data : a geopandas DataFrame
                A geopandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Originally based on Wong, David WS. "Spatial indices of segregation." Urban studies 30.3 (1993): 559-572.

    However, Tivadar, Mihai. "OasisR: An R Package to Bring Some Order to the World of Segregation Measurement." Journal of Statistical Software 89.1 (2019): 1-39.
    points out that in Wong’s original there is an issue with the formula which is an extra division by 2 in the spatial interaction component.
    This function follows the formula present in the first Appendix of Tivadar, Mihai. "OasisR: An R Package to Bring Some Order to the World of Segregation Measurement." Journal of Statistical Software 89.1 (2019): 1-39.

    References: :cite:`wong1993spatial` and :cite:`tivadar2019oasisr`.

    """
    if type(standardize) is not bool:
        raise TypeError("std is not a boolean object")

    D = _dissim(data, group_pop_var, total_pop_var)[0]

    # If a unit has zero population, the group of interest frequency is zero
    data = data.assign(
        pi=np.where(
            data[total_pop_var] == 0, 0, data[group_pop_var] / data[total_pop_var]
        )
    )

    if not standardize:
        cij = _return_length_weighted_w(data).sparse.toarray()
    else:
        cij = _return_length_weighted_w(data).sparse.toarray()
        cij = cij / cij.sum()

    peri = data.length
    ai = data.area

    aux_sum = np.add(
        np.array(list((peri / ai))),
        np.array(list((peri / ai))).reshape((len(list((peri / ai))), 1)),
    )

    max_pa = max(peri / ai)

    num = np.multiply(
        np.multiply(manhattan_distances(data[["pi"]]), cij), aux_sum
    ).sum()
    den = 2 * max_pa

    PARD = D - (num / den)
    PARD

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return PARD, core_data


class PARDissim(SingleGroupIndex, SpatialExplicitIndex):
    """Perimeter-Area Ratio Dissimilarity Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    standardize : boolean
        A condition for row standardisation of the weights matrices. If True, the values of cij in the formulas gets row standardized.
        For the sake of comparison, the seg R package of Hong, Seong-Yun, David O'Sullivan, and Yukio Sadahiro. "Implementing spatial segregation measures in R." PloS one 9.11 (2014): e113767.
        works by default with row standardization.

    Attributes
    ----------
    statistic : float
        PARDissim Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Originally based on Wong, David WS. "Spatial indices of segregation." Urban studies 30.3 (1993): 559-572.

    However, Tivadar, Mihai. "OasisR: An R Package to Bring Some Order to the World of Segregation Measurement." Journal of Statistical Software 89.1 (2019): 1-39.
    points out that in Wong’s original there is an issue with the formula which is an extra division by 2 in the spatial interaction component.
    This function follows the formula present in the first Appendix of Tivadar, Mihai. "OasisR: An R Package to Bring Some Order to the World of Segregation Measurement." Journal of Statistical Software 89.1 (2019): 1-39.

    References: :cite:`wong1993spatial` and :cite:`tivadar2019oasisr`.
    """

    def __init__(
        self, data, group_pop_var, total_pop_var, w=None, standardize=True, **kwargs
    ):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        SpatialExplicitIndex.__init__(self,)
        self.standardize = standardize
        aux = _perimeter_area_ratio_spatial_dissim(
            self.data, self.group_pop_var, self.total_pop_var, self.standardize
        )

        self.statistic = aux[0]
        self.core_data = aux[1]
        self._function = _perimeter_area_ratio_spatial_dissim



"""Calculate street network-based segregation measures."""

__author__ = "Elijah Knaap <elijah.knaap@ucr.edu> Renan X. Cortes <renanc@ucr.edu> and Sergio J. Rey <sergio.rey@ucr.edu>"

import geopandas as gpd
import pandas as pd
from tqdm.auto import tqdm


def _reproject_osm_nodes(nodes_df, input_crs, output_crs):
    #  take original x,y coordinates and convert into geopandas.Series, then reproject
    nodes = gpd.points_from_xy(x=nodes_df.x, y=nodes_df.y, crs=input_crs).to_crs(
        output_crs
    )
    #  convert to dataframe and recreate the x and y cols
    nodes = gpd.GeoDataFrame(index=nodes_df.index, geometry=nodes)
    nodes["x"] = nodes.centroid.x
    nodes["y"] = nodes.centroid.y
    return nodes


def calc_access(
    geodataframe,
    network,
    distance=2000,
    decay="linear",
    variables=None,
    precompute=True,
    return_node_data=False,
):
    """Calculate access to population groups.

    Parameters
    ----------
    geodataframe : geopandas.GeoDataFrame
        geodataframe with demographic data
    network : pandana.Network
        pandana.Network instance. This is likely created with `get_osm_network`
        or via helper functions from OSMnet or UrbanAccess.
    distance : int
        maximum distance to consider `accessible` (the default is 2000).
    decay : str
        decay type pandana should use "linear", "exp", or "flat"
        (which means no decay). The default is "linear".
    variables : list
        list of variable names present on gdf that should be calculated
    precompute: bool (default True)
        whether pandana should precompute the distance matrix. It can only be
        precomputed once, so If you plan to pass the same network to this
        function several times, you should set precompute=False for later runs
    return_node_data : bool, default is False
        Whether to return nodel-level accessibility data or to trim output to
        the same geometries as the input. Default is the latter.

    Returns
    -------
    pandas.DataFrame
        DataFrame with two columns, `total_population` and `group_population`
        which represent the total number of each group that can be reached
        within the supplied `distance` parameter. The DataFrame is indexed
        on node_ids

    """
    if not decay:
        raise Exception("You must pass a decay function such as `linear`")
    if precompute:
        network.precompute(distance)

    geodataframe["node_ids"] = network.get_node_ids(
        geodataframe.centroid.x, geodataframe.centroid.y
    )
    access = []
    for variable in variables:
        network.set(
            geodataframe.node_ids, variable=geodataframe[variable], name=variable
        )

        access_pop = network.aggregate(distance, type="sum", decay=decay, name=variable)

        access.append(access_pop)
    access = pd.DataFrame(dict(zip(variables, access)))
    if return_node_data:
        return access.round(0)
    access = geodataframe[["node_ids", geodataframe.geometry.name]].merge(
        access, right_index=True, left_on="node_ids", how="left"
    )

    return access.dropna()


def compute_travel_cost_matrix(origins, destinations, network, reindex_name=None):
    """Compute a shortest path matrix from a pandana network

    Parameters
    ----------
    origins : geopandas.GeoDataFrame
        the set of origin geometries. If polygon input, the function will use their centroids
    destinations : geopandas.GeoDataFrame
        the set of destination geometries. If polygon input, the function will use their centroids
    network : pandana.Network
        Initialized pandana Network object holding a travel network for a study region
    reindex_name : str, optional
        Name of column on the origin/destinatation dataframe that holds unique index values
        If none (default), the index of the pandana Network node will be used

    Returns
    -------
    pandas.DataFrame
        an origin-destination cost matrix. Rows are origin indices, columns are destination indices,
        and values are shortest network path cost between the two
    """
    origins = origins.copy()
    destinations = destinations.copy()

    #  Note: these are not necessarily "OSM" ids, they're just the identifiers for each  node.
    #  with an integrated ped/transit network, these could be bus stops...
    origins["osm_ids"] = network.get_node_ids(origins.centroid.x, origins.centroid.y)

    destinations["osm_ids"] = network.get_node_ids(
        destinations.centroid.x, destinations.centroid.y
    )

    ods = {}

    with tqdm(total=len(origins["osm_ids"])) as pbar:
        for origin in origins["osm_ids"]:
            ods[f"{origin}"] = network.shortest_path_lengths(
                [origin] * len(origins), destinations["osm_ids"]
            )
            pbar.update(1)

    if reindex_name:
        df = pd.DataFrame(ods, index=origins[reindex_name])
        df.columns = df.index
    else:
        df = pd.DataFrame(ods, index=origins)

    return df


def project_network(network, output_crs=None, input_crs=4326):
    """Reproject a pandana.Network object into another coordinate system

    Parameters
    ----------
    network : pandana.Network
        an instantiated pandana Network object
    input_crs : int, optional
        the coordinate system used in the Network.node_df dataframe. Typically
        these data are collected in Lon/Lat, so the default 4326
    output_crs : int, str, or pyproj.crs.CRS, required
        EPSG code or pyproj.crs.CRS object of the output coordinate system

    Returns
    -------
    pandana.Network
        an initialized pandana.Network with 'x' and y' values represented
        by coordinates in the specified CRS
    """
    try:
        import pandana as pdna
    except ImportError:
        raise ImportError(
            "You need pandana to work with segregation's network module\n"
            "You can install them with  `pip install urbanaccess pandana` "
            "or `conda install -c udst pandana urbanaccess`"
        )

    assert output_crs, "You must provide an output CRS"

    #  take original x,y coordinates and convert into geopandas.Series, then reproject
    nodes = _reproject_osm_nodes(network.nodes_df, input_crs, output_crs)

    #  reinstantiate the network (needs to rebuild the tree)
    net = pdna.Network(
        node_x=nodes["x"],
        node_y=nodes["y"],
        edge_from=network.edges_df["from"],
        edge_to=network.edges_df["to"],
        edge_weights=network.edges_df[network.impedance_names],
        twoway=network._twoway,
    )
    return net



import matplotlib.pyplot as plt
import numpy as np


def plot_cdf(group_share1, group_share2, label1='', label2=''):
    """Plot CDF for two series.

    Convenience function for comparing inequality between two series by
    plotting their CDFs on the same graph

    Parameters
    ----------
    group_share1 : pd.Series
        pandas series with variable of interest.
    group_share2 : pd.Series
        pandas series with variable of interest.
    label1 : str
        legend label for first series
    label2: str
        legend label for second series

    Returns
    -------
    type
        matplotlib Figure.

    """
    plt.step(group_share1.sort_values(),
             group_share1.rank(pct=True).sort_values(),
             label=label1)
    plt.step(group_share2.sort_values(),
             group_share2.rank(pct=True).sort_values(),
             label=label2)
    if (label1 != '' or label2 != ''):
        plt.legend()
    plt.show()
    fig = plt.gcf()
    return fig


def lorenz(X):
    """Plot lorenx curve.

    Parameters
    ----------
    X : pandas.Series
        series of values to plot.

    Returns
    -------
    type
        matplotlib.Figure

    """

    fig, ax = plt.subplots(figsize=[6, 6])

    X = X.sort_values()
    X_lorenz = X.cumsum() / X.sum()
    ax.scatter(np.arange(X_lorenz.size) / (X_lorenz.size - 1),
               X_lorenz,
               marker='x',
               color='navy',
               s=10)
    ax.plot([0, 1], [0, 1], color='k', linewidth=1)
    return ax



"""
Decomposition Segregation based Metrics
"""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Elijah Knaap <elijah.knaap@ucr.edu>, and Sergio J. Rey <sergio.rey@ucr.edu>"


import warnings
import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd
from segregation.inference.comparative import _generate_counterfactual

# Including old and new api in __all__ so users can use both

__all__ = ["DecomposeSegregation"]

# The Deprecation calls of the classes are located in the end of this script #


def _decompose_segregation(index1, index2, counterfactual_approach="composition"):
    """Decompose segregation differences into spatial and attribute components.

    Given two segregation indices of the same type, use Shapley decomposition
    to measure whether the differences between index measures arise from
    differences in spatial structure or population structure

    Parameters
    ----------
    index1 : segregation.SegIndex class
        First SegIndex class to compare.
    index2 : segregation.SegIndex class
        Second SegIndex class to compare.
    counterfactual_approach : str, one of
                              ["composition", "share", "dual_composition"]
        The technique used to generate the counterfactual population
        distributions.

    Returns
    -------
    tuple
        (shapley spatial component,
         shapley attribute component,
         core data of index1,
         core data of index2,
         data with counterfactual variables for index1,
         data with counterfactual variables for index2)

    """
    df1 = index1.data.copy()
    df2 = index2.data.copy()

    assert (
        index1._function == index2._function
    ), "Segregation indices must be of the same type"

    counterfac_df1, counterfac_df2 = _generate_counterfactual(
        df1,
        df2,
        index1.group_pop_var,
        index1.total_pop_var,
        index2.group_pop_var,
        index2.total_pop_var,
        counterfactual_approach=counterfactual_approach,
    )

    seg_func = index1._function

    # index for spatial 1, attribute 1
    G_S1_A1 = index1.statistic

    # index for spatial 2, attribute 2
    G_S2_A2 = index2.statistic

    # index for spatial 1 attribute 2 (counterfactual population for structure 1)
    G_S1_A2 = seg_func(
        counterfac_df1, "counterfactual_group_pop", "counterfactual_total_pop"
    )[0]

    # index for spatial 2 attribute 1 (counterfactual population for structure 2)
    G_S2_A1 = seg_func(
        counterfac_df2, "counterfactual_group_pop", "counterfactual_total_pop"
    )[0]

    # take the average difference in spatial structure, holding attributes constant
    C_S = 1 / 2 * (G_S1_A1 - G_S2_A1 + G_S1_A2 - G_S2_A2)

    # take the average difference in attributes, holding spatial structure constant
    C_A = 1 / 2 * (G_S1_A1 - G_S1_A2 + G_S2_A1 - G_S2_A2)

    results = {"s1_a1": G_S1_A1, "s1_a2": G_S1_A2, "s2_a1": G_S2_A1, "s2_a2": G_S2_A2}

    return (
        C_S,
        C_A,
        df1,
        df2,
        counterfac_df1,
        counterfac_df2,
        counterfactual_approach,
        results,
    )


class DecomposeSegregation:
    """Decompose segregation differences into spatial and attribute components.

    Given two segregation indices of the same type, use Shapley decomposition
    to measure whether the differences between index measures arise from
    differences in spatial structure or population structure

    Parameters
    ----------
    index1 : segregation.SegIndex class
        First SegIndex class to compare.
    index2 : segregation.SegIndex class
        Second SegIndex class to compare.
    counterfactual_approach : str, one of {"composition", "share", "dual_composition"}
        The technique used to generate the counterfactual population
        distributions.

    Attributes
    ----------
    c_s : float
        Shapley's Spatial Component of the decomposition
    c_a : float
        Shapley's Attribute Component of the decomposition
    indices : dict
        Dictionary of index values for all four combinations of spatial/attribute data


    """

    def __init__(self, index1, index2, counterfactual_approach="composition"):
        """Initialize class."""
        aux = _decompose_segregation(index1, index2, counterfactual_approach)

        self.c_s = aux[0]
        self.c_a = aux[1]
        self._df1 = aux[2]
        self._df2 = aux[3]
        self._counterfac_df1 = aux[4]
        self._counterfac_df2 = aux[5]
        self._counterfactual_approach = aux[6]
        self.indices = aux[7]

    def plot(
        self,
        plot_type="cdfs",
        figsize=None,
        city_a=None,
        city_b=None,
        cmap="OrRd",
        scheme="equalinterval",
        k=10,
        suptitle_size=16,
        title_size=12,
        savefig=None,
        dpi=300,
    ):
        """Plot maps or CDFs of urban contexts used in calculating the Decomposition class.

        Parameters
        ----------
        plot_type : str, {'cdfs, 'maps'}
            which type of plot to generate. Options include `cdfs` and `maps` by default "cdfs"
        figsize : tuple, optional
            figsize parameter passed to matplotlib.pyplot
        city_a : str, optional
            Name of the first "city" to be used in plotting. If None, defaults to 'City A'
        city_b : str, optional
            Name of the second "city" to be used in plotting. If None, defaults to 'City B'
        cmap : str, optional
            matplotlib colormap used to shade the map, by default "OrRd"
        scheme : str, optional
            pysal.mapclassify classification scheme used to shade the map, by default "equalinterval"
        k : int, optional
            number of classes in pysal.mapclassify classification scheme, by default 10
        suptitle_size : int, optional
            size parameter passed to `matplotlib.Figure.suptitle`, by default 16
        title_size : int, optional
            size parameter passed to `matplotlib.Axes.set_title`, by default 12
        savefig : str, optional
            Location to save the figure if desired. If None, fig will not be saved
        dpi : int, optional
            dpi parameter passed to matplotlib.pyplot, by default 300

        Returns
        -------
        None
            Generates a new matplotlib.Figure instance and optionally saves to disk
        """
        if not city_a:
            city_a = "City A"
        if not city_b:
            city_b = "City B"

        if plot_type == "cdfs":
            if not figsize:
                figsize = (10, 10)
            fig, ax = plt.subplots(figsize=figsize)
            plt.suptitle(
                f"Decomposing differences between\n{city_a} and {city_b}",
                size=suptitle_size,
            )
            plt.title(
                f"Spatial Component = {round(self.c_s, 3)}, Attribute Component: {round(self.c_a, 3)}",
                size=title_size,
            )

            temp_a = self._counterfac_df1.copy()
            temp_a["Location"] = city_a
            temp_b = self._counterfac_df2.copy()
            temp_b["Location"] = city_b
            df = pd.concat([temp_a, temp_b]).reset_index()

            if self._counterfactual_approach == "composition":
                sns.ecdfplot(data=df, x="group_composition", hue="Location", ax=ax)
                return ax

            elif self._counterfactual_approach == "share":
                f = sns.ecdfplot(data=df, x="share", hue="Location", ax=ax)
                return f

            elif self._counterfactual_approach == "dual_composition":
                df["compl"] = 1 - df.group_composition
                f = sns.ecdfplot(data=df, x="group_composition", hue="Location", ax=ax)
                f2 = sns.ecdfplot(data=df, x="compl", hue="Location", ax=ax)
            if savefig:
                plt.savefig(savefig, dpi=dpi)

        if plot_type == "maps":
            if not figsize:
                figsize = (20, 20)
            fig, axs = plt.subplots(2, 2, figsize=figsize)
            plt.suptitle(
                f"Decomposing differences between\n{city_a} and {city_b}",
                size=suptitle_size,
            )
            plt.title(
                f"Spatial Component = {round(self.c_s, 3)}, Attribute Component: {round(self.c_a, 3)}"
            )

            # Original First Context (Upper Left)
            self._counterfac_df1.plot(
                column="group_composition",
                cmap=cmap,
                legend=True,
                scheme=scheme,
                k=k,
                ax=axs[0, 0],
            )
            axs[0, 0].set_title(
                f"{city_a}\nOriginal Composition", fontdict={"fontsize": title_size}
            )
            axs[0, 0].axis("off")

            # Counterfactual First Context (Bottom Left)
            self._counterfac_df1.plot(
                column="counterfactual_composition",
                cmap=cmap,
                scheme=scheme,
                k=k,
                legend=True,
                ax=axs[1, 0],
            )
            axs[1, 0].set_title(
                f"{city_a}\nCounterfactual Composition",
                fontdict={"fontsize": title_size},
            )
            axs[1, 0].axis("off")

            # Counterfactual Second Context (Upper Right)
            self._counterfac_df2.plot(
                column="counterfactual_composition",
                cmap=cmap,
                scheme=scheme,
                k=k,
                legend=True,
                ax=axs[0, 1],
            )
            axs[0, 1].set_title(
                f"{city_b}\nCounterfactual Composition",
                fontdict={"fontsize": title_size},
            )
            axs[0, 1].axis("off")

            # Original Second Context (Bottom Right)
            self._counterfac_df2.plot(
                column="group_composition",
                cmap=cmap,
                scheme=scheme,
                k=k,
                legend=True,
                ax=axs[1, 1],
            )
            axs[1, 1].set_title(
                f"{city_b}\nOriginal Composition", fontdict={"fontsize": title_size}
            )
            axs[1, 1].axis("off")
            if savefig:
                plt.savefig(savefig, dpi=dpi)
            return axs



"""Relative Clustering Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
import pandas as pd
from ..util import generate_distance_matrix
from .._base import SingleGroupIndex, SpatialExplicitIndex


def _relative_clustering(data, group_pop_var, total_pop_var, alpha=0.6, beta=0.5):
    """Calculate Relative Clustering index.

    Parameters
    ----------
    data          : a geopandas DataFrame with a geometry column.
    group_pop_var : string
                    The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
                    The name of variable in data that contains the total population of the unit
    alpha         : float
                    A parameter that estimates the extent of the proximity within the same unit. Default value is 0.6
    beta          : float
                    A parameter that estimates the extent of the proximity within the same unit. Default value is 0.5

    Returns
    ----------
    statistic : float
                Relative Clustering Index
    core_data : a geopandas DataFrame
                A geopandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    The pairwise distance between unit i and itself is (alpha * area_of_unit_i) ^ beta.

    Reference: :cite:`massey1988dimensions`.

    """
    if alpha < 0:
        raise ValueError("alpha must be greater than zero.")

    if beta < 0:
        raise ValueError("beta must be greater than zero.")

    data = data.assign(
        xi=data[group_pop_var], yi=data[total_pop_var] - data[group_pop_var]
    )

    X = data.xi.sum()
    Y = data.yi.sum()

    dist = generate_distance_matrix(data)

    np.fill_diagonal(dist, val=np.exp(-((alpha * data.area.values) ** (beta))))

    c = 1 - dist.copy()  # proximity matrix
    Pxx = (data.xi.values * data.xi.values * c).sum() / (X ** 2)
    Pyy = (data.yi.values * data.yi.values * c).sum() / (Y ** 2)
    RCL = (Pxx / Pyy) - 1

    if np.isnan(RCL):
        raise ValueError(
            "It not possible to determine the distance between, at least, one pair of units. This is probably due to the magnitude of the number of the centroids. We recommend to reproject the geopandas DataFrame."
        )

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return RCL, core_data


class RelativeClustering(SingleGroupIndex, SpatialExplicitIndex):
    """Relative Clustering Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    alpha  : float
        A parameter that estimates the extent of the proximity within the same unit. Default value is 0.6
    beta : float
        A parameter that estimates the extent of the proximity within the same unit. Default value is 0.5

    Attributes
    ----------
    statistic : float
        Relative Clustering Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    The pairwise distance between unit i and itself is (alpha * area_of_unit_i) ^ beta.

    Reference: :cite:`massey1988dimensions`.
    """

    def __init__(
        self, data, group_pop_var, total_pop_var, alpha=0.6, beta=0.5, **kwargs,
    ):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        SpatialExplicitIndex.__init__(self,)
        self.alpha = alpha
        self.beta = beta
        aux = _relative_clustering(
            self.data, self.group_pop_var, self.total_pop_var, self.alpha, self.beta,
        )

        self.statistic = aux[0]
        self.core_data = aux[1]
        self._function = _relative_clustering



"""Multigroup Diversity index"""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
from geopandas import GeoDataFrame

from .._base import MultiGroupIndex, SpatialImplicitIndex

np.seterr(divide="ignore", invalid="ignore")


def _multi_diversity(data, groups, normalized=False):
    """Calculate of Multigroup Diveristy Index

    Parameters
    ----------
    data   : a pandas DataFrame
    groups : list of strings.
             The variables names in data of the groups of interest of the analysis.

    Returns
    -------

    statistic  : float
                 Multigroup Diversity Index
    core_data  : a pandas DataFrame
                 A pandas DataFrame that contains the columns used to perform the estimate.
    normalized : bool. Default is False.
                 Wheter the resulting index will be divided by its maximum (natural log of the number of groups)

    Notes
    -----
    Based on Reardon, Sean F., and Glenn Firebaugh. "Measures of multigroup segregation." Sociological methodology 32.1 (2002): 33-67 and Theil, Henry. "Statistical decomposition analysis; with applications in the social and administrative sciences". No. 04; HA33, T4.. 1972.

    This is also know as Theil's Entropy Index (Equation 2 of page 37 of Reardon, Sean F., and Glenn Firebaugh. "Measures of multigroup segregation." Sociological methodology 32.1 (2002): 33-67)

    High diversity means less segregation.

    Reference: :cite:`reardon2002measures`.

    """

    core_data = data[groups]
    df = np.array(core_data)

    Pk = df.sum(axis=0) / df.sum()

    E = -(Pk * np.log(Pk)).sum()

    if normalized:
        K = df.shape[1]
        E = E / np.log(K)
    if isinstance(data, GeoDataFrame):
        core_data = data[[data.geometry.name]].join(core_data)
    return E, core_data, groups


class MultiDiversity(MultiGroupIndex, SpatialImplicitIndex):
    """Multigroup Diversity Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object
    normalized : bool. Default is False.
        Whether the resulting index will be divided by its maximum (natural log of the number of groups)
    
    Attributes
    ----------
    statistic : float
        Multigroup Dissimilarity Index value
    core_data : a pandas DataFrame
        DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Reardon, Sean F., and Glenn Firebaugh. "Measures of multigroup segregation." Sociological methodology 32.1 (2002): 33-67 and Theil, Henry. "Statistical decomposition analysis; with applications in the social and administrative sciences". No. 04; HA33, T4.. 1972.

    This is also know as Theil's Entropy Index (Equation 2 of page 37 of Reardon, Sean F., and Glenn Firebaugh. "Measures of multigroup segregation." Sociological methodology 32.1 (2002): 33-67)

    High diversity means less segregation.

    Reference: :cite:`reardon2002measures`.
    """

    def __init__(
        self,
        data,
        groups,
        w=None,
        normalized=False,
        network=None,
        distance=None,
        decay=None,
        precompute=None,
        function='triangular',
        **kwargs
    ):
        """Init."""
        MultiGroupIndex.__init__(self, data, groups)
        self.normalized = normalized
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(self, w, network, distance, decay, function, precompute)
        aux = _multi_diversity(self.data, self.groups, normalized=self.normalized)

        self.statistic = aux[0]
        self.data = aux[1]
        self.groups = aux[2]
        self._function = _multi_diversity



"""Absolute Centralization Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
from scipy.ndimage import shift

from .._base import SingleGroupIndex, SpatialExplicitIndex


def _absolute_centralization(
    data, group_pop_var, total_pop_var, center="mean", metric="euclidean"
):
    """Calculation of Absolute Centralization index.

    Parameters
    ----------
    data : a geopandas DataFrame with a geometry column.
    group_pop_var : string
        The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
        The name of variable in data that contains the total population of the unit
    center : string, two-dimension values (tuple, list, array) or integer.
        This defines what is considered to be the center of the spatial context under study.
        If string, this can be set to:

            "mean": the center longitude/latitude is the mean of longitudes/latitudes of all units.
            "median": the center longitude/latitude is the median of longitudes/latitudes of all units.
            "population_weighted_mean": the center longitude/latitude is the mean of longitudes/latitudes of all units weighted by the total population.
            "largest_population": the center longitude/latitude is the centroid of the unit with largest total population. If there is a tie in the maximum population, the mean of all coordinates will be taken.

        If tuple, list or array, this argument should be the coordinates of the desired center assuming longitude as first value and latitude second value. Therefore, in the form (longitude, latitude), if tuple, or [longitude, latitude] if list or numpy array.

        If integer, the center will be the centroid of the polygon from data corresponding to the integer interpreted as index.
        For example, if `center = 0` the centroid of the first row of data is used as center, if `center = 1` the second row will be used, and so on.
    metric : string. Can be 'euclidean' or 'haversine'. Default is 'euclidean'.
        The metric used for the distance between spatial units.
        If the projection of the CRS of the geopandas DataFrame field is in degrees, this should be set to 'haversine'.

    Returns
    ----------
    statistic     : float
        Absolute Centralization Index
    core_data     : a geopandas DataFrame
        A geopandas DataFrame that contains the columns used to perform the estimate.
    center_values : list
        The center, in the form [longitude, latitude], values used for the calculation of the centralization distances.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    A discussion of defining the center in this function can be found in https://github.com/pysal/segregation/issues/18.

    Reference: :cite:`massey1988dimensions`.
    """

    if metric not in ["euclidean", "haversine"]:
        raise ValueError("metric must one of 'euclidean', 'haversine'")

    x = np.array(data[group_pop_var])
    t = np.array(data[total_pop_var])

    if any(t < x):
        raise ValueError(
            "Group of interest population must equal or lower than the total population of the units."
        )

    area = np.array(data.area)

    c_lons = np.array(data.centroid.x)
    c_lats = np.array(data.centroid.y)

    if isinstance(center, str):
        if center not in [
            "mean",
            "median",
            "population_weighted_mean",
            "largest_population",
        ]:
            raise ValueError(
                "The center string must one of 'mean', 'median', 'population_weighted_mean', 'largest_population'"
            )

        if center == "mean":
            center_lon = c_lons.mean()
            center_lat = c_lats.mean()

        if center == "median":
            center_lon = np.median(c_lons)
            center_lat = np.median(c_lats)

        if center == "population_weighted_mean":
            center_lon = np.average(c_lons, weights=t)
            center_lat = np.average(c_lats, weights=t)

        if center == "largest_population":
            center_lon = c_lons[np.where(t == t.max())].mean()
            center_lat = c_lats[np.where(t == t.max())].mean()

    if (
        isinstance(center, tuple)
        or isinstance(center, list)
        or isinstance(center, np.ndarray)
    ):
        if np.array(center).shape != (2,):
            raise ValueError("The center tuple/list/array must have length 2.")

        center_lon = center[0]
        center_lat = center[1]

    if isinstance(center, int):
        if (center > len(data) - 1) or (center < 0):
            raise ValueError("The center index must by in the range of data.")

        center_lon = data.iloc[[center]].centroid.x.values[0]
        center_lat = data.iloc[[center]].centroid.y.values[0]

    X = x.sum()
    A = area.sum()

    dlon = c_lons - center_lon
    dlat = c_lats - center_lat

    if metric == "euclidean":
        center_dist = np.sqrt((dlon) ** 2 + (dlat) ** 2)

    if metric == "haversine":
        center_dist = 2 * np.arcsin(
            np.sqrt(
                np.sin(dlat / 2) ** 2
                + np.cos(center_lat) * np.cos(c_lats) * np.sin(dlon / 2) ** 2
            )
        )

    if np.isnan(center_dist).sum() > 0:
        raise ValueError(
            "It not possible to determine the center distance for, at least, one unit. This is probably due to the magnitude of the number of the centroids. We recommend to reproject the geopandas DataFrame."
        )

    asc_ind = center_dist.argsort()

    Xi = np.cumsum(x[asc_ind]) / X
    Ai = np.cumsum(area[asc_ind]) / A

    ACE = np.nansum(shift(Xi, 1, cval=np.NaN) * Ai) - np.nansum(
        Xi * shift(Ai, 1, cval=np.NaN)
    )

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    center_values = [center_lon, center_lat]

    return ACE, core_data, center_values


class AbsoluteCentralization(SingleGroupIndex, SpatialExplicitIndex):
    """Absolute Centralization Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    center  : string, two-dimension values (tuple, list, array) or integer.
        This defines what is considered to be the center of the spatial context under study.
        If string, this can be set to:

            "mean": the center longitude/latitude is the mean of longitudes/latitudes of all units.
            "median": the center longitude/latitude is the median of longitudes/latitudes of all units.
            "population_weighted_mean": the center longitude/latitude is the mean of longitudes/latitudes of all units weighted by the total population.
            "largest_population": the center longitude/latitude is the centroid of the unit with largest total population. If there is a tie in the maximum population, the mean of all coordinates will be taken.
    metric : str
        The metric used for the distance between spatial units.
        If the projection of the CRS of the geopandas DataFrame field is in degrees, this should be set to 'haversine'.


    Attributes
    ----------
    statistic : float
        AbsoluteCentralization Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    A discussion of defining the center in this function can be found in https://github.com/pysal/segregation/issues/18.

    Reference: :cite:`massey1988dimensions`.
    """

    def __init__(
        self,
        data,
        group_pop_var,
        total_pop_var,
        center="mean",
        metric="euclidean",
        **kwargs,
    ):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        SpatialExplicitIndex.__init__(self,)
        self.center = center
        self.metric = metric
        aux = _absolute_centralization(
            self.data, self.group_pop_var, self.total_pop_var, self.center, self.metric,
        )

        self.statistic = aux[0]
        self.core_data = aux[1]
        self.center_values = aux[2]
        self._function = _absolute_centralization



"""Modified Gini Segregation Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import geopandas as gpd
import numpy as np
import pandas as pd
from .._base import SingleGroupIndex, SpatialImplicitIndex
from .gini import _gini_seg
from tqdm.auto import tqdm
from joblib import Parallel, delayed
import multiprocessing
from warnings import warn


def _modified_gini(
    data,
    group_pop_var,
    total_pop_var,
    iterations=500,
    backend="threading",
    n_jobs=-1,
    seed=None,
):
    """Calculate Modified Gini index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame
        Dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : string
        Variable containing the population count of the group of interest
    total_pop_var : string
        Variable in data that contains the total population count of the unit
    iterations : int
        The number of iterations the evaluate average classic dissimilarity under eveness.
        Default value is 500.
    n_jobs : int
        [Optional. Default=-1] Number of processes to run in parallel. If -1,
        this is set to the number of CPUs available
    backend : str {'loky', 'threading'}
        backend to pass into joblib's Parallel constructor.
    seed : int
        random seed passed to np.random inside the parallelization constructor to return
        consistent results


    Returns
    ----------
    statistic : float
        Modified Gini Index (Gini from Carrington and Troske (1997))
    data : pandas.DataFrame
        pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Carrington, William J., and Kenneth R. Troske. "On measuring segregation in samples with small units." Journal of Business & Economic Statistics 15.4 (1997): 402-409.

    Reference: :cite:`carrington1997measuring`.
    """
    if not seed:
        seed = np.random.randint(
            0, 10e6
        )  # is there a better practice for this? I think joblib will fail if None passed
    if n_jobs == -1:
        n_jobs = multiprocessing.cpu_count()

    D = _gini_seg(data, group_pop_var, total_pop_var)[0]

    x = data[group_pop_var].values
    t = data[total_pop_var].values.astype(int)

    p_null = x.sum() / t.sum()

    # Ds = np.empty(iterations)

    def _gen_estimate(i):
        n_retries = 5
        try:
            while n_retries > 0:
                data = i[0]
                n = i[1]
                p = i[2]
                np.random.seed(i[3])
                freq_sim = np.random.binomial(
                    n=n,
                    p=p,
                    size=(1, data.shape[0]),
                ).tolist()[0]
                data[group_pop_var] = freq_sim
                aux = _gini_seg(data, group_pop_var, total_pop_var)[0]
                return aux

        except ValueError:
            warn("Simulator generated invalid data. Redrawing")
            n_retries -= 1

    Ds = pd.Series(
        Parallel(n_jobs=n_jobs, backend=backend)(
            delayed(_gen_estimate)(
                (
                    data,
                    np.array([t.tolist()]),
                    np.array([[p_null] * data.shape[0]]),
                    seed,
                )
            )
            for i in range(iterations)
        )
    )

    D_star = Ds.mean()

    if D >= D_star:
        Dct = (D - D_star) / (1 - D_star)
    else:
        Dct = (D - D_star) / D_star

    if not isinstance(data, gpd.GeoDataFrame):
        core_data = data[[group_pop_var, total_pop_var]]

    else:
        core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return Dct, core_data


class ModifiedGini(SingleGroupIndex, SpatialImplicitIndex):
    """Modified Gini Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
        Modified Gini Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    Reference: :cite:`massey1988dimensions`.
    """

    def __init__(
        self,
        data,
        group_pop_var,
        total_pop_var,
        iterations=500,
        w=None,
        network=None,
        distance=None,
        decay="linear",
        function="triangular",
        precompute=None,
        backend="threading",
        n_jobs=-1,
        **kwargs
    ):
        """Init."""

        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(
                self, w, network, distance, decay, function, precompute
            )
        aux = _modified_gini(
            self.data,
            self.group_pop_var,
            self.total_pop_var,
            iterations,
            backend=backend,
            n_jobs=n_jobs,
        )

        self.statistic = aux[0]
        self.data = aux[1]
        self._function = _modified_gini



import unittest
from libpysal.examples import load_example
import geopandas as gpd
import numpy as np
from segregation.singlegroup import BiasCorrectedDissim


class Bias_Corrected_Dissim_Tester(unittest.TestCase):
    def test_Bias_Corrected_Dissim(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        df = s_map[['geometry', 'HISP', 'TOT_POP']]
        np.random.seed(1234)
        index = BiasCorrectedDissim(df, 'HISP', 'TOT_POP')
        np.testing.assert_almost_equal(index.statistic, 0.32136474449360836, decimal = 3)


if __name__ == '__main__':
    unittest.main()



import unittest

import geopandas as gpd
import numpy as np
from libpysal.examples import load_example
from segregation.local import LocalRelativeCentralization


class Local_Relative_Centralization_Tester(unittest.TestCase):
    def test_Local_Relative_Centralization(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        df = s_map[["geometry", "BLACK", "TOT_POP"]]
        index = LocalRelativeCentralization(df, "BLACK", "TOT_POP")
        np.testing.assert_almost_equal(
            index.statistics[0:10],
            np.array(
                [
                    0.03443055,
                    -0.29063264,
                    -0.19110976,
                    0.24978919,
                    0.01252249,
                    0.61152941,
                    0.78917647,
                    0.53129412,
                    0.04436346,
                    -0.20216325,
                ]
            ),
        )


if __name__ == "__main__":
    unittest.main()



"""Absolute Clustering Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
import pandas as pd
from ..util import generate_distance_matrix
from .._base import SingleGroupIndex, SpatialExplicitIndex


def _absolute_clustering(data, group_pop_var, total_pop_var, alpha=0.6, beta=0.5):
    """Calculation of Absolute Clustering index

    Parameters
    ----------
    data : geopandas.GeoDataFrame
        a GeoDataFrame with a geometry column
    group_pop_var : string
        The name of variable in data that contains the population size of the
        group of interest
    total_pop_var : string
        The name of variable in data that contains the total population of the
        unit
    alpha : float
        A parameter that estimates the extent of the proximity within the same
        unit. Default value is 0.6
    beta : float
        A parameter that estimates the extent of the proximity within the same
        unit. Default value is 0.5

    Returns
    ----------
    statistic : float
        Absolute Clustering Index
    core_data : a geopandas DataFrame
        A geopandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    The pairwise distance between unit i and itself is (alpha * area_of_unit_i) ^ beta.

    Reference: :cite:`massey1988dimensions`.

    """
    if alpha < 0:
        raise ValueError("alpha must be greater than zero.")

    if beta < 0:
        raise ValueError("beta must be greater than zero.")

    X = data[group_pop_var].values.sum()

    x = data[group_pop_var].values
    t = data[total_pop_var].values
    n = len(data)

    dist = generate_distance_matrix(data)
    np.fill_diagonal(dist, val=np.exp(-((alpha * data.area.values) ** (beta))))

    c = 1 - dist.copy()  # proximity matrix
    ACL = ((((x / X) * (c * x).sum(axis=1)).sum()) - ((X / n ** 2) * c.sum())) / (
        (((x / X) * (c * t).sum(axis=1)).sum()) - ((X / n ** 2) * c.sum())
    )

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return ACL, core_data


class AbsoluteClustering(SingleGroupIndex, SpatialExplicitIndex):
    """Absolute Clustering Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of
        interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    alpha : float
        A parameter that estimates the extent of the proximity within the same unit.
        Default value is 0.6
    beta : float
        A parameter that estimates the extent of the proximity within the same unit.
        Default value is 0.5


    Attributes
    ----------
    statistic : float
        AbsolutecClustering Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    The pairwise distance between unit i and itself is (alpha * area_of_unit_i) ^ beta.

    Reference: :cite:`massey1988dimensions`.
    """

    def __init__(
        self, data, group_pop_var, total_pop_var, alpha=0.6, beta=0.5, **kwargs,
    ):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        SpatialExplicitIndex.__init__(self,)
        self.alpha = alpha
        self.beta = beta
        aux = _absolute_clustering(
            self.data, self.group_pop_var, self.total_pop_var, self.alpha, self.beta,
        )

        self.statistic = aux[0]
        self.core_data = aux[1]
        self._function = _absolute_clustering



"""Gini Segregation Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import geopandas as gpd
import numpy as np

from .._base import SingleGroupIndex, SpatialImplicitIndex



try:
    from numba import njit, jit, prange, boolean
except (ImportError, ModuleNotFoundError):

    def jit(*dec_args, **dec_kwargs):
        """
        decorator mimicking numba.jit
        """

        def intercepted_function(f, *f_args, **f_kwargs):
            return f

        return intercepted_function

    njit = jit

    prange = range
    boolean = bool

@njit(parallel=True, fastmath=True,)
def _gini_vecp(pi: np.ndarray, ti: np.ndarray):
    """Memory efficient calculation of Gini

    Parameters
    ----------
    pi : np.ndarray
        area minority population counts
    ti : np.ndarray
        area total population counts

    Returns
    ----------
    
    implicit: float
             Gini coefficient
    """


    n = ti.shape[0]
    num = np.zeros(1)
    T = ti.sum()
    P = pi.sum() / T
    pi = np.where(ti == 0, 0, pi / ti)
    T = ti.sum()
    for i in prange(n-1):
        num += (ti[i] * ti[i+1:] * np.abs(pi[i] - pi[i+1:])).sum()
    num *= 2
    den = (2 * T * T * P * (1-P))
    return (num / den)[0]

    

def _gini_seg(data, group_pop_var, total_pop_var):
    """Calculate Gini segregation index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame
        Dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : string
        Variable containing the population count of the group of interest
    total_pop_var : string
        Variable in data that contains the total population count of the unit

    Returns
    ----------
    statistic : float
        MinMax index statistic value
    core_data : pandas.DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    Reference: :cite:`massey1988dimensions`.
    """

    # If a unit has zero population, the group of interest frequency is zero
    data = data.assign(
        ti=data[total_pop_var],
        pi=np.where(
            data[total_pop_var] == 0, 0, data[group_pop_var] / data[total_pop_var]
        ),
    )

    pi = data[group_pop_var].values
    ti = data[total_pop_var].values
    G = _gini_vecp(pi, ti)

    if not isinstance(data, gpd.GeoDataFrame):
        data = data[[group_pop_var, total_pop_var]]

    else:
        data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return G, data

class Gini(SingleGroupIndex, SpatialImplicitIndex):
    """Gini Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
                Gini Index
    core_data : a pandas DataFrame
                A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    Reference: :cite:`massey1988dimensions`.
    """

    def __init__(
        self,
        data,
        group_pop_var,
        total_pop_var,
        w=None,
        network=None,
        distance=None,
        decay=None,
        function="triangular",
        precompute=None,
        **kwargs
    ):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(
                self, w, network, distance, decay, function, precompute
            )
        aux = _gini_seg(self.data, self.group_pop_var, self.total_pop_var)

        self.statistic = aux[0]
        self.data = aux[1]
        self._function = _gini_seg



import unittest
from libpysal.examples import load_example
import geopandas as gpd
import numpy as np
from segregation.local import MultiLocalEntropy


class Multi_Local_Entropy_Tester(unittest.TestCase):
    def test_Multi_Local_Entropy(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        groups_list = ['WHITE', 'BLACK', 'ASIAN','HISP']
        df = s_map[groups_list]
        index = MultiLocalEntropy(df, groups_list)
        np.testing.assert_almost_equal(index.statistics[0:10], np.array([0.24765538, 0.40474253, 0.50900607, 0.21433739, 0.16148146,
																		 0.21454691, 0.08889013, 0.08132889, 0.06782401, 0.18127186]))


if __name__ == '__main__':
    unittest.main()



import unittest

import geopandas as gpd
import numpy as np
from libpysal.examples import load_example
from segregation.local import MultiLocalDiversity


class Multi_Local_Diversity_Tester(unittest.TestCase):
    def test_Multi_Local_Diversity(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        groups_list = ["WHITE", "BLACK", "ASIAN", "HISP"]
        df = s_map[groups_list]
        index = MultiLocalDiversity(df, groups_list)
        np.testing.assert_almost_equal(
            index.statistics[0:10],
            np.array(
                [
                    0.34332326,
                    0.56109229,
                    0.70563225,
                    0.29713472,
                    0.22386084,
                    0.29742517,
                    0.12322789,
                    0.11274579,
                    0.09402405,
                    0.25129616,
                ]
            ),
        )


if __name__ == "__main__":
    unittest.main()



"""Bias Corrected Dissimilarity index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import geopandas as gpd
import numpy as np
import pandas as pd

from .._base import SingleGroupIndex, SpatialImplicitIndex
from .dissim import _dissim


def _bias_corrected_dissim(data, group_pop_var, total_pop_var, B=500):
    """
    Calculation of Bias Corrected Dissimilarity index

    Parameters
    ----------

    data : pandas.DataFrame or geopandas.GeoDataFrame
        DataFrame holding necessary data
    group_pop_var : string
        The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
        The name of variable in data that contains the total population of the unit
    B : int
       The number of iterations to calculate Dissimilarity simulating randomness with multinomial distributions. Default value is 500.

    Returns
    ----------
    statistic : float
        Dissimilarity with Bias-Correction (bias correction from Allen, Rebecca et al. (2015))
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Allen, Rebecca, et al. "More reliable inference for the dissimilarity index of segregation." The econometrics journal 18.1 (2015): 40-66.

    Reference: :cite:`allen2015more`.
    """
    assert type(B) is int, "B must be an integer"

    assert B > 1, "B must be greater than 1."

    D = _dissim(data, group_pop_var, total_pop_var)[0]

    x = np.array(data[group_pop_var])
    t = np.array(data[total_pop_var])

    other_group_pop = t - x

    # Group 0: minority group
    p0_i = x / x.sum()
    n0 = x.sum()
    sim0 = np.random.multinomial(n0, p0_i, size=B)

    # Group 1: complement group
    p1_i = other_group_pop / other_group_pop.sum()
    n1 = other_group_pop.sum()
    sim1 = np.random.multinomial(n1, p1_i, size=B)

    Dbcs = np.empty(B)
    for i in np.array(range(B)):
        data_aux = {
            "simul_group": sim0[i].tolist(),
            "simul_tot": (sim0[i] + sim1[i]).tolist(),
        }
        df_aux = pd.DataFrame.from_dict(data_aux)
        Dbcs[i] = _dissim(df_aux, "simul_group", "simul_tot")[0]

    Db = Dbcs.mean()

    Dbc = 2 * D - Db
    Dbc  # It expected to be lower than D, because D is upwarded biased

    if isinstance(data, gpd.GeoDataFrame):
        core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    else:
        core_data = data[[group_pop_var, total_pop_var]]

    return Dbc, core_data


class BiasCorrectedDissim(SingleGroupIndex, SpatialImplicitIndex):
    """Bias Corrected Dissimilarity Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    B : int
       The number of iterations to calculate Dissimilarity simulating randomness with multinomial distributions. Default value is 500.
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
                BiasCorrectedDissim Index
    core_data : a pandas DataFrame
                A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Carrington, William J., and Kenneth R. Troske. "On measuring segregation in samples with small units." Journal of Business & Economic Statistics 15.4 (1997): 402-409.

    Reference: :cite:`carrington1997measuring`.
    """

    def __init__(
        self,
        data,
        group_pop_var,
        total_pop_var,
        B=500,
        w=None,
        network=None,
        distance=None,
        decay=None,
        precompute=None,
        function="triangular",
        **kwargs
    ):
        """Init."""

        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(
                self, w, network, distance, decay, function, precompute
            )
        self.B = B
        aux = _bias_corrected_dissim(
            self.data, self.group_pop_var, self.total_pop_var, self.B
        )

        self.statistic = aux[0]
        self.data = aux[1]
        self._function = _bias_corrected_dissim



import unittest
from libpysal.examples import load_example
import geopandas as gpd
import numpy as np
from segregation.local import MultiLocalSimpsonConcentration


class Multi_Local_Simpson_Concentration_Tester(unittest.TestCase):
    def test_Multi_Local_Simpson_Concentration(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        groups_list = ['WHITE', 'BLACK', 'ASIAN','HISP']
        df = s_map[groups_list]
        index = MultiLocalSimpsonConcentration(df, groups_list)
        np.testing.assert_almost_equal(index.statistics[0:10], np.array([0.84564007, 0.66608405, 0.50090253, 0.8700551 , 0.90194944,
																		 0.86871822, 0.95552644, 0.9601067 , 0.96276946, 0.88241452]))


if __name__ == '__main__':
    unittest.main()



"""Delta Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np

from .._base import SingleGroupIndex, SpatialExplicitIndex


def _delta(data, group_pop_var, total_pop_var):
    """Calculate Delta index.

    Parameters
    ----------
    data          : a geopandas DataFrame with a geometry column.
    group_pop_var : string
                    The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
                    The name of variable in data that contains the total population of the unit

    Returns
    ----------
    statistic : float
                Delta Index
    core_data : a geopandas DataFrame
                A geopandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    Reference: :cite:`massey1988dimensions`.

    """
    x = np.array(data[group_pop_var])
    t = np.array(data[total_pop_var])

    if any(t < x):
        raise ValueError(
            "Group of interest population must equal or lower than the total population of the units."
        )

    area = np.array(data.area)

    X = x.sum()
    A = area.sum()

    DEL = 1 / 2 * abs(x / X - area / A).sum()

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return DEL, core_data


class Delta(SingleGroupIndex, SpatialExplicitIndex):
    """Delta Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population

    Attributes
    ----------
    statistic : float
        Delta Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.


    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    Reference: :cite:`massey1988dimensions`.
    """

    def __init__(
        self, data, group_pop_var, total_pop_var, **kwargs,
    ):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        SpatialExplicitIndex.__init__(self,)
        aux = _delta(self.data, self.group_pop_var, self.total_pop_var,)

        self.statistic = aux[0]
        self.core_data = aux[1]
        self._function = _delta



import unittest
from libpysal.examples import load_example
import geopandas as gpd
import numpy as np
from segregation.singlegroup import MinMax


class SpatialMinMax_Tester(unittest.TestCase):
    def test_SpatialMinMax(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        df = s_map[['geometry', 'HISP', 'TOT_POP']]
        df = df.to_crs(df.estimate_utm_crs())
        index = MinMax(df, 'HISP', 'TOT_POP', distance=2000, function='triangular')
        np.testing.assert_almost_equal(index.statistic, 0.4524336967483127, decimal=4)


if __name__ == '__main__':
    unittest.main()



from libpysal.examples import load_example
import geopandas as gpd
import pandana as pdna
from segregation.network import calc_access


import quilt3 as q3
b = q3.Bucket("s3://spatial-ucr")
b.fetch("osm/metro_networks_8k/40900.h5", "./40900.h5")

def test_calc_access():
    variables = ['WHITE', 'BLACK', 'ASIAN', 'HISP']
    s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
    df = s_map[['FIPS', 'geometry'] + variables]
    df = df[df.FIPS.str.startswith('06061')]
    df = df[(df.centroid.x < -121) & (df.centroid.y < 38.85)]
    df.crs = {'init': 'epsg:4326'}
    df[variables] = df[variables].astype(float)
    test_net = pdna.Network.from_hdf5('40900.h5')

    acc = calc_access(df, test_net, distance=1., variables=variables)
    assert acc.WHITE.sum() > 100



"""Spatial Proximity Profile Segregation Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
from libpysal.weights import Queen
from scipy.sparse.csgraph import floyd_warshall
from numba import njit
from .._base import SingleGroupIndex, SpatialExplicitIndex


def _spatial_prox_profile(data, group_pop_var, total_pop_var, w, m):
    """Calculate Spatial Proximity Profile.

    Parameters
    ----------
    data : geopandas.GeoDataFrame (required)
        GeoDataFrame with valid geometry column and columns for group population and total population
    group_pop_var : string
        The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
        The name of variable in data that contains the total population of the unit
    w: libpysal.weights.W
        pysal spatial weights object measuring connectivity between geographic units. If nNne, a Queen object will be created
    m : int
        a numeric value indicating the number of thresholds to be used. Default value is 1000.
        A large value of m creates a smoother-looking graph and a more precise spatial proximity profile value but slows down the calculation speed.

    Returns
    ----------
    statistic : float
        Spatial Proximity Index
    core_data : geopandas.GeoDataFrame
        A GeoDataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Hong, Seong-Yun, and Yukio Sadahiro. "Measuring geographic segregation: a graph-based approach." Journal of Geographical Systems 16.2 (2014): 211-231.

    Reference: :cite:`hong2014measuring`.

    """
    # Create the shortest distance path between two pair of units using Shimbel matrix.
    # This step was well discussed in https://github.com/pysal/segregation/issues/5.
    if not w:
        w = Queen.from_dataframe(data)
    delta = floyd_warshall(csgraph=w.sparse, directed=False)
    group_vals = data[group_pop_var].to_numpy()
    total_vals = data[total_pop_var].to_numpy()
    
    grid = np.linspace(0, 1, m)

    @njit(fastmath=True, error_model="numpy")
    def calc(grid):
        def calculate_etat(t):
            g_t_i = np.where(np.divide(group_vals, total_vals) >= t, True, False)
            k = g_t_i.sum()

            # i and j only varies in the units subset within the threshold in eta_t of Hong (2014).
            sub_delta_ij = delta[g_t_i, :][:, g_t_i]

            den = sub_delta_ij.sum()
            eta_t = (k ** 2 - k) / den
            return eta_t

        results = np.empty(len(grid))
        for i, est in enumerate(grid):
            aux = calculate_etat(est)
            results[i] = aux
        return results

    aux = calc(grid)
    aux[aux == np.inf] = 0
    aux[aux == -np.inf] = 0
    curve = np.nan_to_num(aux, 0)

    threshold = data[group_pop_var].sum() / data[total_pop_var].sum()
    SPP = (
        threshold
        - ((curve[grid < threshold]).sum() / m - (curve[grid >= threshold]).sum() / m)
    ) / (1 - threshold)

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return SPP, grid, curve, core_data


class SpatialProxProf(SingleGroupIndex, SpatialExplicitIndex):
    """Spatial Proximity Profile Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    w: libpysal.weights.W
        pysal spatial weights object measuring connectivity between geographic units. If nNne, a Queen object will be created
    m : int
        a numeric value indicating the number of thresholds to be used. Default value is 1000.
        A large value of m creates a smoother-looking graph and a more precise spatial proximity
        profile value but slows down the calculation speed.

    Attributes
    ----------
    statistic : float
        Spatial Prox Profile Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Hong, Seong-Yun, and Yukio Sadahiro. "Measuring geographic segregation: a graph-based approach." Journal of Geographical Systems 16.2 (2014): 211-231.

    Reference: :cite:`hong2014measuring`.
    """

    def __init__(self, data, group_pop_var, total_pop_var, w=None, m=1000, **kwargs):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        SpatialExplicitIndex.__init__(self,)
        self.m = m
        self.w = w
        aux = _spatial_prox_profile(
            self.data, self.group_pop_var, self.total_pop_var, self.w, self.m
        )

        self.statistic = aux[0]
        self.grid = aux[1]
        self.curve = aux[2]
        self.core_data = aux[3]
        self._function = _spatial_prox_profile

    def plot(self):
        """Plot the Spatial Proximity Profile."""
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ImportError("This method relies on importing `matplotlib`")
        graph = plt.scatter(self.grid, self.curve, s=0.1)
        return graph



import numpy as np
import pandas as pd

from scipy.spatial.distance import pdist, squareform
from scipy.special import rel_entr as relative_entropy

from ..network import compute_travel_cost_matrix
from warnings import warn


def compute_divergence_profiles(
    gdf, groups, metric="euclidean", network=None, distance_matrix=None
):
    """
    A segregation metric using Kullback-Leiber (KL) divergence to quantify the
    difference in the population characteristics between (1) an area and (2) the total population.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    metric : str (optional; 'euclidean' by default)
        Distance metric for calculating pairwise distances,
        Accepts any inputs to `scipy.spatial.distance.pdist`.
        Ignored if passing a network or distance matrix
    network: pandana.Network object (optional, None by default)
        A pandana Network object used to compute distance between observations
    distance_matrix: numpy.array (optional; None by default)
        numpy array of distances between observations in the dataset

    Returns
    ----------
    aux : geopandas.GeoDataFrame
        geodataframe of the KL divergence measure, between the aggregated population and the
        total population, will converge to zero for the final row of each
        observation to represent that the total population is covered.
        population_covered : the population count within the aggregated population.
        Returns a concatenated object of Pandas dataframes. Each dataframe contains a
        set of divergence levels between an area and the total population. These areas
        become consecutively larger, starting from a single location and aggregating
        outward from this location, until the area represents the total population.
        Thus, together the divergence levels within a dataframe represent a profile
        of divergence from an area. The concatenated object is the collection of these
        divergence profiles for every areas within the total population.

    """
    # Store the observation index to return with the results
    indices = gdf.index.copy()
    centroids = gdf.geometry.centroid
    df = gdf[groups].values

    coordinates = np.column_stack((centroids.x, centroids.y))

    # If given a pandana network, use shortest network distance, otherwise use scikit
    if network:
        if metric != "network":
            warn(
                f"metric set to {metric} but a pandana.Network object was passed. Using network distances instead"
                "If you wish to use a scipy distance matrix, do not include a `network` argument`"
            )
        dist_matrix = compute_travel_cost_matrix(gdf, gdf, network).values
    elif distance_matrix:
        if metric != "precomputed":
            warn(
                f"metric set to {metric} but a distance_matrix argument was passed. Using precomputed distances instead"
            )
        dist_matrix = distance_matrix
    else:
        dist_matrix = squareform(pdist(coordinates, metric=metric))

    # Preparing list for results
    results = []

    # Loop to calculate KL divergence
    for (i, distances) in enumerate(dist_matrix):

        # Creating the q and r objects
        sorted_indices = np.argsort(distances)
        cumul_pop_by_group = np.cumsum(df[sorted_indices], axis=0)
        obs_cumul_pop = np.sum(cumul_pop_by_group, axis=1)[:, np.newaxis]
        q_cumul_proportions = cumul_pop_by_group / obs_cumul_pop
        total_pop_by_group = np.sum(df, axis=0, keepdims=True)
        total_pop = np.sum(df)
        r_total_proportions = total_pop_by_group / total_pop

        # Input q and r objects into relative entropy (KL divergence) function
        kl_divergence = relative_entropy(q_cumul_proportions, r_total_proportions).sum(
            axis=1
        )

        # Creating an output dataframe
        output = pd.DataFrame().from_dict(
            dict(
                observation=indices[i],
                distance=distances[sorted_indices],
                divergence=kl_divergence,
                population_covered=obs_cumul_pop.sum(axis=1),
            )
        )

        # Append (bring together) all outputs into results list
        results.append(output)

    aux = pd.concat(results)

    return aux



"""CorrelationR Segregation Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import geopandas as gpd
import numpy as np
import pandas as pd

from .._base import SingleGroupIndex, SpatialImplicitIndex


def _correlationr(data, group_pop_var, total_pop_var):
    """Calculation of Correlation Ratio index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame
        Dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : string
        Variable containing the population count of the group of interest
    total_pop_var : string
        Variable in data that contains the total population count of the unit

    Returns
    ----------
    statistic : float
        MinMax index statistic value
    core_data : pandas.DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    Reference: :cite:`massey1988dimensions`.

    """
    x = np.array(data[group_pop_var])
    t = np.array(data[total_pop_var])

    X = x.sum()
    T = t.sum()
    P = X / T

    xPx = np.nansum((x / X) * (x / t))

    V = (xPx - P) / (1 - P)

    if not isinstance(data, gpd.GeoDataFrame):
        core_data = data[[group_pop_var, total_pop_var]]

    else:
        core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return V, core_data


class CorrelationR(SingleGroupIndex, SpatialImplicitIndex):
    """CorrelationR Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
                CorrelationR Index
    core_data : a pandas DataFrame
                A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Massey, Douglas S., and Nancy A. Denton. "The dimensions of residential segregation." Social forces 67.2 (1988): 281-315.

    Reference: :cite:`massey1988dimensions`.
    """

    def __init__(
        self,
        data,
        group_pop_var,
        total_pop_var,
        w=None,
        network=None,
        distance=None,
        decay=None,
        function="triangular",
        precompute=None,
        **kwargs
    ):
        """Init."""

        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(
                self, w, network, distance, decay, function, precompute
            )
        aux = _correlationr(self.data, self.group_pop_var, self.total_pop_var)

        self.statistic = aux[0]
        self.data = aux[1]
        self._function = _correlationr



"""Spatial Dissimilarity Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import libpysal
import numpy as np
from libpysal.weights import Queen

from .._base import SingleGroupIndex, SpatialExplicitIndex
from .dissim import _dissim


def _spatial_dissim(data, group_pop_var, total_pop_var, w=None, standardize=False):
    """Calculate of Spatial Dissimilarity index.

    Parameters
    ----------
    data : a geopandas DataFrame with a geometry column.
    group_pop_var : string
        The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
        The name of variable in data that contains the total population of the unit
    w : W
        A PySAL weights object. If not provided, Queen contiguity matrix is used.
    standardize  : boolean
        A condition for row standardisation of the weights matrices. If True, the values of cij in the formulas gets row standardized.
        For the sake of comparison, the seg R package of Hong, Seong-Yun, David O'Sullivan, and Yukio Sadahiro. "Implementing spatial segregation measures in R." PloS one 9.11 (2014): e113767.
        works by default with row standardization.

    Returns
    ----------
    statistic : float
        Spatial Dissimilarity Index
    core_data : a geopandas DataFrame
        A geopandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Morrill, R. L. (1991) "On the Measure of Geographic Segregation". Geography Research Forum.

    Reference: :cite:`morrill1991measure`.

    """
    if type(standardize) is not bool:
        raise TypeError("std is not a boolean object")

    if w is None:
        w_object = Queen.from_dataframe(data)
    else:
        w_object = w

    if not issubclass(type(w_object), libpysal.weights.W):
        raise TypeError("w is not a PySAL weights object")

    D = _dissim(data, group_pop_var, total_pop_var)[0]

    x = np.array(data[group_pop_var])
    t = np.array(data[total_pop_var])

    # If a unit has zero population, the group of interest frequency is zero
    pi = np.where(t == 0, 0, x / t)

    if not standardize:
        cij = w_object.sparse.toarray()
    else:
        cij = w_object.sparse.toarray()
        cij = cij / cij.sum(axis=1).reshape((cij.shape[0], 1))

    # Inspired in (second solution): https://stackoverflow.com/questions/22720864/efficiently-calculating-a-euclidean-distance-matrix-using-numpy
    # Distance Matrix
    abs_dist = abs(pi[..., np.newaxis] - pi)

    # manhattan_distances used to compute absolute distances
    num = np.multiply(abs_dist, cij).sum()
    den = cij.sum()
    SD = D - num / den
    SD

    core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return SD, core_data


class SpatialDissim(SingleGroupIndex, SpatialExplicitIndex):
    """Spatial Dissimilarity Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    w: libpysal.weights.W
        pysal spatial weights object measuring connectivity between geographic units. If nNne, a Queen object will be created
    standardize : boolean
        A condition for row standardisation of the weights matrices. If True, the values of cij in the formulas gets row standardized.
        For the sake of comparison, the seg R package of Hong, Seong-Yun, David O'Sullivan, and Yukio Sadahiro. "Implementing spatial segregation measures in R." PloS one 9.11 (2014): e113767.
        works by default with row standardization.

    Attributes
    ----------
    statistic : float
        SpatialDissim Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Morrill, R. L. (1991) "On the Measure of Geographic Segregation". Geography Research Forum.

    Reference: :cite:`morrill1991measure`.
    """

    def __init__(
        self, data, group_pop_var, total_pop_var, w=None, standardize=False, **kwargs
    ):
        """Init."""
        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        SpatialExplicitIndex.__init__(self,)
        self.w = w
        self.standardize = standardize
        aux = _spatial_dissim(
            self.data, self.group_pop_var, self.total_pop_var, self.w, self.standardize
        )

        self.statistic = aux[0]
        self.core_data = aux[1]
        self._function = _spatial_dissim



"""Multigroup Information index"""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
from geopandas import GeoDataFrame

from .._base import MultiGroupIndex, SpatialImplicitIndex

np.seterr(divide="ignore", invalid="ignore")


def _multi_information_theory(data, groups):
    """Calculate Multigroup Information Theory index.

    Parameters
    ----------

    data   : a pandas DataFrame

    groups : list of strings.
             The variables names in data of the groups of interest of the analysis.

    Returns
    -------

    statistic : float
                Multigroup Information Theory Index

    core_data : a pandas DataFrame
                A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Reardon, Sean F., and Glenn Firebaugh. "Measures of multigroup segregation." Sociological methodology 32.1 (2002): 33-67.

    Reference: :cite:`reardon2002measures`.

    """
    data = data.copy()
    core_data = data[groups]
    df = np.array(core_data)

    T = df.sum()

    ti = df.sum(axis=1)
    pik = df / ti[:, None]
    Pk = df.sum(axis=0) / df.sum()

    # The natural logarithm is used, but this could be used with any base following Footnote 3 of pg. 37
    # of Reardon, Sean F., and Glenn Firebaugh. "Measures of multigroup segregation." Sociological methodology 32.1 (2002): 33-67.
    E = (Pk * np.log(1 / Pk)).sum()

    MIT = np.nansum(ti[:, None] * pik * np.log(pik / Pk)) / (T * E)
    if isinstance(data, GeoDataFrame):
        core_data = data[[data.geometry.name]].join(core_data)
    return MIT, core_data, groups


class MultiInfoTheory(MultiGroupIndex, SpatialImplicitIndex):
    """Multigroup Information Theory Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
        Multigroup Dissimilarity Index value
    core_data : a pandas DataFrame
        DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Reardon, Sean F., and Glenn Firebaugh. "Measures of multigroup segregation." Sociological methodology 32.1 (2002): 33-67.

    Reference: :cite:`reardon2002measures`.
    """

    def __init__(
        self,
        data,
        groups,
        w=None,
        network=None,
        distance=None,
        decay='linear',
        function='triangular',
        precompute=False,
        **kwargs
    ):
        """Init."""
        MultiGroupIndex.__init__(self, data, groups)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(self, w, network, distance, decay, function, precompute)
        aux = _multi_information_theory(self.data, self.groups)

        self.statistic = aux[0]
        self.data = aux[1]
        self.groups = aux[2]
        self._function = _multi_information_theory



from .multi_dissim import MultiDissim
from .multi_divergence import MultiDivergence
from .multi_diversity import MultiDiversity
from .multi_gini import MultiGini
from .multi_info_theory import MultiInfoTheory
from .multi_norm_exposure import MultiNormExposure
from .multi_relative_diversity import MultiRelativeDiversity
from .multi_squared_coef_var import MultiSquaredCoefVar
from .simpsons_concentration import SimpsonsConcentration
from .simpsons_interaction import SimpsonsInteraction
from .multi_global_distortion import GlobalDistortion



import unittest
from libpysal.examples import load_example
import geopandas as gpd
import numpy as np
from segregation.local import MultiLocalSimpsonInteraction


class Multi_Local_Simpson_Interaction_Tester(unittest.TestCase):
    def test_Multi_Local_Simpson_Interaction(self):
        s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))
        groups_list = ['WHITE', 'BLACK', 'ASIAN','HISP']
        df = s_map[groups_list]
        index = MultiLocalSimpsonInteraction(df, groups_list)
        np.testing.assert_almost_equal(index.statistics[0:10], np.array([0.15435993, 0.33391595, 0.49909747, 0.1299449 , 0.09805056,
																		 0.13128178, 0.04447356, 0.0398933 , 0.03723054, 0.11758548]))


if __name__ == '__main__':
    unittest.main()



from tkinter import N
import geopandas as gpd
import numpy as np
from libpysal.examples import load_example
from segregation.batch import (
    batch_compute_multigroup,
    batch_compute_singlegroup,
    batch_multiscalar_singlegroup,
    batch_multiscalar_multigroup,
)

s_map = gpd.read_file(load_example("Sacramento1").get_path("sacramentot2.shp"))


def test_batch_single():
    np.random.seed(1234)
    fit = batch_compute_singlegroup(
        s_map.to_crs(s_map.estimate_utm_crs()),
        group_pop_var="HISP",
        total_pop_var="TOT_POP",
        distance=2000,
        center="mean",
        function="triangular",
        seed=1234,
        backend='loky'
        # loky is slower but more robust in testing

    )
    np.testing.assert_array_almost_equal(
        fit.Statistic,
        [
            0.6905,
            0.061,
            0.8511,
            0.1218,
            0.2922,
            0.2639,
            0.1243,
            0.0746,
            0.8043,
            0.22,
            0.2924,
            0.86,
            0.2,
            0.0767,
            0.3932,
            0.7717,
            0.2283,
            0.4524,
            0.2839,
            0.3827,
            0.3113,
            -0.1034,
            0.52,
            0.1264,
            0.2612,
            0.2285,
            1.0468,
        ],
        decimal=2,
    )


def test_batch_multi():
    np.random.seed(1234)
    mfit = batch_compute_multigroup(
        s_map.to_crs(s_map.estimate_utm_crs()),
        distance=2000,
        groups=["HISP", "BLACK", "WHITE"],
    )
    np.testing.assert_array_almost_equal(
        mfit.Statistic,
        [   11.78 ,
            0.37768411,
            0.11294892,
            0.78242435,
            0.50485431,
            0.14435762,
            0.15922567,
            0.13971496,
            0.11631982,
            0.55281779,
            0.44718221,
        ],
        decimal=3,
    )


def test_batch_multiscalar_multi():
    np.random.seed(1234)

    mfit = batch_multiscalar_multigroup(
        s_map.to_crs(s_map.estimate_utm_crs()),
        distances=[500, 1000],
        groups=["HISP", "BLACK", "WHITE"],
    )
    assert mfit.shape == (3, 10)


def test_batch_multiscalar_single():
    mfit = batch_multiscalar_singlegroup(
        s_map.to_crs(s_map.estimate_utm_crs()),
        distances=[500, 1000],
        group_pop_var="HISP",
        total_pop_var="TOT_POP",
        seed=1234,
        backend='loky'
        # loky is slower but more robust in testing
    )
    assert mfit.shape == (3, 13)




"""
Useful functions for segregation metrics
"""

__author__ = "Levi Wolf <levi.john.wolf@gmail.com>, Renan X. Cortes <renanc@ucr.edu>, and Eli Knaap <ek@knaaptime.com>"


import numpy as np
import pandas as pd
import warnings
from sklearn.metrics.pairwise import euclidean_distances


def generate_distance_matrix(data):
    """Generate a pairwise Euclidean distance matrix from the rows in a geodataframe.

    Parameters
    ----------
    data : geopandas.GeoDataFrame
        A geodataframe with a projected coordinate system

    Returns
    -------
    numpy.array
        distance matrix
    """
    assert data.crs.is_projected, "This function can only operate on euclidean coordinates. Please reproject the geodataframe"
    w = euclidean_distances(
        pd.DataFrame({"x": data.centroid.x.values, "y": data.centroid.y.values})
    )
    w = w / w.sum(axis=1)
    dist = np.exp(-w)
    return dist


def _nan_handle(df):
    """Check if dataframe has nan values.
    Raise an informative error.
    """
    if str(type(df)) == "<class 'geopandas.geodataframe.GeoDataFrame'>":
        values = df.loc[:, df.columns != df._geometry_column_name].values
    else:
        values = df.values

    if np.any(np.isnan(values)):
        warnings.warn(
            "There are NAs present in the input data. NAs should be handled (e.g. dropping or replacing them with values) before using this function."
        )

    return df



"""Compute multiscalar segregation profiles."""

import warnings

import numpy as np
import pandas as pd
from libpysal.weights import Kernel
from pyproj.crs import CRS


def compute_multiscalar_profile(
    gdf,
    segregation_index=None,
    groups=None,
    group_pop_var=None,
    total_pop_var=None,
    distances=None,
    network=None,
    decay="linear",
    function="triangular",
    precompute=True,
    **kwargs
):
    """Compute multiscalar segregation profile.

    This function calculates several Spatial Information Theory indices with
    increasing distance parameters.

    Parameters
    ----------
    gdf : geopandas.GeoDataFrame
        geodataframe with rows as observations and columns as population
        variables. Note that if using a network distance, the coordinate
        system for this gdf should be 4326. If using euclidian distance,
        this must be projected into planar coordinates like state plane or UTM.
    segregation_index : SpatialImplicit SegregationIndex Class
        a class from the library such as MultiInformationTheory, or MinMax
    groups : list
        list of population groups for calculating multigroup indices
    group_pop_var : str
        name of population group on gdf for calculating single group indices
    total_pop_var : str
        bame of total population on gdf for calculating single group indices
    distances : list
        list of floats representing bandwidth distances that define a local
        environment.
    network : pandana.Network (optional)
        A pandana.Network likely created with
        `segregation.network.get_osm_network`.
    decay : str (optional)
        decay type to be used in pandana accessibility calculation
        options are {'linear', 'exp', 'flat'}. The default is 'linear'.
    function: 'str' (optional)
        which weighting function should be passed to libpysal.weights.Kernel
        must be one of: 'triangular','uniform','quadratic','quartic','gaussian'
    precompute: bool
        Whether the pandana.Network instance should precompute the range
        queries. This is True by default
    **kwargs : dict
        additional keyword arguments passed to each index (e.g. for setting a random
        seed in indices like ModifiedGini or ModifiedDissm)


    Returns
    -------
    pandas.Series
        Series with distances as index and index statistics as values

    Notes
    -----
    Based on Sean F. Reardon, Stephen A. Matthews, David O’Sullivan, Barrett A. Lee, Glenn Firebaugh, Chad R. Farrell, & Kendra Bischoff. (2008). The Geographic Scale of Metropolitan Racial Segregation. Demography, 45(3), 489–514. https://doi.org/10.1353/dem.0.0019.

    Reference: :cite:`reardon2008`.

    """
    if not segregation_index:
        raise ValueError("You must pass a segregation SpatialImplicit Index Class")
    gdf = gdf.copy()
    indices = {}

    if groups:
        gdf[groups] = gdf[groups].astype(float)
        indices[0] = segregation_index(gdf, groups=groups, **kwargs).statistic
    elif group_pop_var:
        indices[0] = segregation_index(
            gdf, group_pop_var=group_pop_var, total_pop_var=total_pop_var, **kwargs
        ).statistic

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        if network:
            if not gdf.crs.equals(CRS(4326)):
                gdf = gdf.to_crs(epsg=4326)
            if precompute:
                maxdist = max(distances)
                network.precompute(maxdist)
            for distance in distances:
                distance = np.float(distance)
                if group_pop_var:
                    idx = segregation_index(
                        gdf,
                        group_pop_var=group_pop_var,
                        total_pop_var=total_pop_var,
                        network=network,
                        decay=decay,
                        distance=distance,
                        precompute=False,
                        **kwargs
                    )
                elif groups:
                    idx = segregation_index(
                        gdf,
                        groups=groups,
                        network=network,
                        decay=decay,
                        distance=distance,
                        precompute=False,
                        **kwargs
                    )

                indices[distance] = idx.statistic
        else:
            for distance in distances:
                w = Kernel.from_dataframe(gdf, bandwidth=distance, function=function)
                if group_pop_var:
                    idx = segregation_index(
                        gdf,
                        group_pop_var=group_pop_var,
                        total_pop_var=total_pop_var,
                        w=w,
                        **kwargs
                    )
                else:
                    idx = segregation_index(gdf, groups, w=w, **kwargs)
                indices[distance] = idx.statistic
        series = pd.Series(indices, name=str(type(idx)).split(".")[-1][:-2])
        series.index.name = "distance"
        return series



"""Multigroup Simpson's Concentration index"""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import numpy as np
from geopandas import GeoDataFrame

from .._base import MultiGroupIndex, SpatialImplicitIndex

np.seterr(divide="ignore", invalid="ignore")


def _simpsons_concentration(data, groups):
    """
    Calculation of Simpson's Concentration index

    Parameters
    ----------

    data   : a pandas DataFrame

    groups : list of strings.
             The variables names in data of the groups of interest of the analysis.

    Returns
    -------

    statistic  : float
                 Simpson's Concentration Index

    core_data  : a pandas DataFrame
                 A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Simpson, Edward H. "Measurement of diversity." nature 163.4148 (1949): 688.

    Simpson's concentration index (Lambda) can be simply interpreted as the probability that two individuals chosen at random and independently from the population will be found to belong to the same group.

    Higher values means higher segregation.

    Simpson's Concentration + Simpson's Interaction = 1

    Reference: :cite:`simpson1949measurement`.

    """

    core_data = data[groups]
    df = np.array(core_data)

    Pk = df.sum(axis=0) / df.sum()

    Lambda = (Pk * Pk).sum()
    if isinstance(data, GeoDataFrame):
        core_data = data[[data.geometry.name]].join(core_data)
    return Lambda, core_data, groups


class SimpsonsConcentration(MultiGroupIndex, SpatialImplicitIndex):
    """Simpsons Concentration Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
        Multigroup Dissimilarity Index value
    core_data : a pandas DataFrame
        DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Reardon, Sean F., and Glenn Firebaugh. "Measures of multigroup segregation." Sociological methodology 32.1 (2002): 33-67.

    Reference: :cite:`reardon2002measures`.
    """

    def __init__(
        self,
        data,
        groups,
        w=None,
        network=None,
        distance=None,
        decay=None,
        precompute=None,
        function='triangular',
        **kwargs
    ):
        """Init."""
        MultiGroupIndex.__init__(self, data, groups)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(self, w, network, distance, decay, function, precompute)
        aux = _simpsons_concentration(self.data, self.groups)

        self.statistic = aux[0]
        self.data = aux[1]
        self.groups = aux[2]
        self._function = _simpsons_concentration



import geopandas as gpd

from .._base import MultiGroupIndex, SpatialExplicitIndex
from ..dynamics import compute_divergence_profiles


def _global_distortion(
    gdf, groups, network=None, metric="euclidean", distance_matrix=None, normalize=False
):
    """
    A segregation metric, using Kullback-Leiber (KL) divergence to quantify the
    difference in the population characteristics between (1) an area and (2) the total population.

    This function utilises the methodology proposed in
    Olteanu et al. (2019): 'Segregation through the multiscalar lens'. Which can be
    found here: https://doi.org/10.1073/pnas.1900192116

    Arguments
    ----------
    gdf : geopandas.GeoDataFrame
        geodataframe with group population counts (not percentages) to be included in the analysis.
    groups : list
        list of columns on gdf that contain population counts of interest
    metric : str (optional; 'euclidean' by default)
        Distance metric for calculating pairwise distances,
        Accepts any inputs to `scipy.spatial.distance.pdist`.
        Ignored if passing a network or distance matrix
    network: pandana.Network object (optional)
        A pandana Network object used to compute distance between observations
    distance_matrix: numpy.array
        numpy array of distances between observations in the dataset
    normalization:
        NOT YET IMPLEMENTED


    Returns
    ----------
    gdf: geopands.GeoDataFrame
        a geodataframe of input data used to compute the statistic
    statistic : float
        the global distortion index

    """
    # Store the observation index to return with the results
    geoms = gdf[gdf.geometry.name]
    df = gdf[groups].values

    total_pop = df.sum().sum()

    aux = compute_divergence_profiles(
        gdf=gdf,
        groups=groups,
        network=network,
        metric=metric,
        distance_matrix=distance_matrix,
    )

    #  this yeilds distortion coefficients
    aux = aux.groupby("observation").sum()[["divergence"]]

    if normalize:
        raise Exception("Not yet implemented")
        # Need to write a routine to determine the scaling factor... From the paper:

        # the maximum distortion coefficient in a theoretical extreme case of segregation.
        # Theoretically, the maximal-segregation distortion coefficient is achieved when sorting
        # the k groups into k ghettos, ordered by sizes, and then computing the coefficient for
        # the most isolated person in the smallest group

    # the global multiplies the population at each location by the distortion coefficient they experience
    aux["coefs"] = aux["divergence"] * df.sum(axis=1)
    stat = (1 / total_pop) * aux["coefs"].sum()

    out = gpd.GeoDataFrame(gdf, columns=groups, geometry=geoms, crs=geoms.crs)
    out["weighted_distortion"] = aux["coefs"]

    return stat, out


class GlobalDistortion(MultiGroupIndex, SpatialExplicitIndex):
    """Multigroup Global Distortion Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    groups : list, required
        list of columns on dataframe holding population totals for each group
    metric : str (optional; 'euclidean' by default)
        Distance metric for calculating pairwise distances,
        Accepts any inputs to `scipy.spatial.distance.pdist`.
        Ignored if passing a network or distance matrix
    network: pandana.Network object (optional, None by default)
        A pandana Network object used to compute distance between observations
    distance_matrix: numpy.array (optional; None by default)
        numpy array of distances between observations in the dataset
    normalization: bool
        NOT YET IMPLEMENTED

    Attributes
    ----------
    statistics : pandas.Series
        KL Divergence coefficients
    core_data : a pandas DataFrame
        DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Bézenac, C., Clark, W. A. V., Olteanu, M., & Randon‐Furling, J. (2022). Measuring and Visualizing Patterns
    of Ethnic Concentration: The Role of Distortion Coefficients. Geographical Analysis, 54(1), 173–196.
    https://doi.org/10.1111/gean.12271

    Reference: :cite:`debezenac2021`.
    """

    def __init__(
        self,
        data,
        groups=None,
        metric="euclidean",
        network=None,
        distance_matrix=None,
        normalize=False,
        **kwargs
    ):
        """Init."""

        MultiGroupIndex.__init__(self, data, groups)
        SpatialExplicitIndex.__init__(self)

        aux = _global_distortion(
            self.data,
            self.groups,
            network=network,
            metric=metric,
            normalize=normalize,
            distance_matrix=distance_matrix,
        )

        self.statistic = aux[0]
        self.data = aux[1]
        self._function = _global_distortion



"""Density-Corrected Dissim Segregation Index."""

__author__ = "Renan X. Cortes <renanc@ucr.edu>, Sergio J. Rey <sergio.rey@ucr.edu> and Elijah Knaap <elijah.knaap@ucr.edu>"

import geopandas as gpd
import numpy as np
import pandas as pd
from scipy.optimize import minimize
from warnings import warn

from .._base import SingleGroupIndex, SpatialImplicitIndex


# Constructing function that returns $n(\hat{\theta}_j)$
def _return_optimal_theta(theta_j):
    def fold_norm(x):
        try:
            from rvlib import Normal
            norm = Normal(0, 1)

        except ImportError:
            warn("Unable to import `rvlib`. Falling back to slower scipy sampler")
            from scipy.stats import norm

        y = (-1) * (norm.pdf(x - theta_j) + norm.pdf(x + theta_j))
        return y

    initial_guesses = np.array(0)
    res = minimize(
        fold_norm, initial_guesses, method="nelder-mead", options={"xatol": 1e-5}
    )
    return res.final_simplex[0][1][0]


def _density_corrected_dissim(
    data,
    group_pop_var,
    total_pop_var,
):
    """Calculate Density Corrected Dissimilarity index.

    Parameters
    ----------
    data :  pandas.DataFrame
        DataFrame storing necessary data
    group_pop_var : string
        The name of variable in data that contains the population size of the group of interest
    total_pop_var : string
        The name of variable in data that contains the total population of the unit
    xtol : float
        The degree of tolerance in the optimization process of returning optimal theta_j

    Returns
    ----------
    statistic : float
        Dissimilarity with Density-Correction (density correction from Allen, Rebecca et al. (2015))
    core_data : pandas.DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Allen, Rebecca, et al. "More reliable inference for the dissimilarity index of segregation." The econometrics journal 18.1 (2015): 40-66.

    Reference: :cite:`allen2015more`.
    """
    g = np.array(data[group_pop_var])
    t = np.array(data[total_pop_var])

    other_group_pop = t - g

    # Group 0: minority group
    p0_i = g / g.sum()
    n0 = g.sum()

    # Group 1: complement group
    p1_i = other_group_pop / other_group_pop.sum()
    n1 = other_group_pop.sum()

    sigma_hat_j = np.sqrt(((p1_i * (1 - p1_i)) / n1) + ((p0_i * (1 - p0_i)) / n0))
    theta_hat_j = abs(p1_i - p0_i) / sigma_hat_j

    optimal_thetas = pd.Series(data=theta_hat_j).apply(_return_optimal_theta)

    Ddc = np.multiply(sigma_hat_j, optimal_thetas).sum() / 2

    if not isinstance(data, gpd.GeoDataFrame):
        core_data = data[[group_pop_var, total_pop_var]]

    else:
        core_data = data[[group_pop_var, total_pop_var, data.geometry.name]]

    return Ddc, core_data


class DensityCorrectedDissim(SingleGroupIndex, SpatialImplicitIndex):
    """Density Corrected Dissimilarity Index.

    Parameters
    ----------
    data : pandas.DataFrame or geopandas.GeoDataFrame, required
        dataframe or geodataframe if spatial index holding data for location of interest
    group_pop_var : str, required
        name of column on dataframe holding population totals for focal group
    total_pop_var : str, required
        name of column on dataframe holding total overall population
    w : libpysal.weights.KernelW, optional
        lipysal spatial kernel weights object used to define an egohood
    network : pandana.Network
        pandana Network object representing the study area
    distance : int
        Maximum distance (in units of geodataframe CRS) to consider the extent of the egohood
    decay : str
        type of decay function to apply. Options include
    precompute : bool
        Whether to precompute the pandana Network object

    Attributes
    ----------
    statistic : float
        Segregation Index
    core_data : a pandas DataFrame
        A pandas DataFrame that contains the columns used to perform the estimate.

    Notes
    -----
    Based on Allen, Rebecca, et al. "More reliable inference for the dissimilarity index of segregation." The econometrics journal 18.1 (2015): 40-66.

    Reference: :cite:`allen2015more`.
    """

    def __init__(
        self,
        data,
        group_pop_var,
        total_pop_var,
        w=None,
        network=None,
        distance=None,
        decay="linear",
        precompute=None,
        function="triangular",
        **kwargs
    ):
        """Init."""

        SingleGroupIndex.__init__(self, data, group_pop_var, total_pop_var)
        if any([w, network, distance]):
            SpatialImplicitIndex.__init__(
                self, w, network, distance, decay, function, precompute
            )
        aux = _density_corrected_dissim(
            self.data, self.group_pop_var, self.total_pop_var
        )

        self.statistic = aux[0]
        self.data = aux[1]
        self._function = _density_corrected_dissim
